import datetime
import difflib
import grp
import json
import logging
import os
import os.path
import pwd
import re
import shutil
import sys
import tarfile
import tempfile
import time
from io import BytesIO
from xml.dom.minidom import parse

try:
    import distro

    no_distro_package = False
except ImportError:
    no_distro_package = True
    import platform

# TODO remove, deprecated
try:
    import clufter.command_manager
    import clufter.facts
    import clufter.filter_manager
    import clufter.format_manager

    no_clufter = False
except ImportError:
    no_clufter = True

from pcs import (
    alert,
    cluster,
    constraint,
    quorum,
    settings,
    status,
    stonith,
    usage,
    utils,
)
from pcs.cli.common import middleware
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.output import (
    INDENT_STEP,
    smart_wrap_text,
)
from pcs.cli.constraint import command as constraint_command
from pcs.cli.nvset import nvset_dto_list_to_lines
from pcs.cli.reports import process_library_reports
from pcs.cli.reports.output import warn
from pcs.cli.resource.output import (
    ResourcesConfigurationFacade,
    resources_to_text,
)
from pcs.common.reports import constraints as constraints_reports
from pcs.common.str_tools import indent
from pcs.lib.commands import quorum as lib_quorum
from pcs.lib.errors import LibraryError
from pcs.lib.node import get_existing_nodes_names

# pylint: disable=too-many-branches, too-many-locals, too-many-statements


def config_show(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file, when getting cluster name on remote node (corosync.conf
        doesn't exist)
      * --corosync_conf - corosync.conf file
    """
    modifiers.ensure_only_supported("-f", "--corosync_conf")
    if argv:
        raise CmdLineInputError()
    print("Cluster Name: %s" % utils.getClusterName())
    status.nodes_status(lib, ["config"], modifiers.get_subset("-f"))
    print()
    print("\n".join(_config_show_cib_lines(lib)))
    if (
        utils.hasCorosyncConf()
        and not modifiers.is_specified("-f")
        and not modifiers.is_specified("--corosync_conf")
    ):
        cluster.cluster_uidgid(
            lib, [], modifiers.get_subset(), silent_list=True
        )
    if modifiers.is_specified("--corosync_conf") or utils.hasCorosyncConf():
        print()
        print("Quorum:")
        try:
            config = lib_quorum.get_config(utils.get_lib_env())
            print("\n".join(indent(quorum.quorum_config_to_str(config))))
        except LibraryError as e:
            process_library_reports(e.args)


def _config_show_cib_lines(lib):
    """
    Commandline options:
      * -f - CIB file
    """
    # update of pcs_options will change output of constraint show and
    # displaying resources and operations defaults
    utils.pcs_options["--full"] = 1
    # get latest modifiers object after updating pcs_options
    modifiers = utils.get_input_modifiers()
    cib_dom = utils.get_cib_dom()

    resources_facade = ResourcesConfigurationFacade.from_resources_dto(
        lib.resource.get_configured_resources()
    )

    all_lines = []

    all_lines.append("Resources:")
    all_lines.extend(
        smart_wrap_text(
            indent(
                resources_to_text(resources_facade.filter_stonith(False)),
                indent_step=INDENT_STEP,
            )
        )
    )
    all_lines.append("")
    all_lines.append("Stonith Devices:")
    all_lines.extend(
        smart_wrap_text(
            indent(
                resources_to_text(resources_facade.filter_stonith(True)),
                indent_step=INDENT_STEP,
            )
        )
    )
    all_lines.append("Fencing Levels:")
    levels_lines = stonith.stonith_level_config_to_str(
        lib.fencing_topology.get_config()
    )
    if levels_lines:
        all_lines.extend(indent(levels_lines, indent_step=2))

    all_lines.append("")
    constraints_element = cib_dom.getElementsByTagName("constraints")[0]
    all_lines.extend(
        constraint.location_lines(
            constraints_element,
            showDetail=True,
            show_expired=True,
            verify_expiration=False,
        )
    )
    all_lines.extend(
        constraint_command.config_cmd(
            "Ordering Constraints:",
            lib.constraint_order.config,
            constraints_reports.order_plain,
            modifiers.get_subset("-f", "--full"),
        )
    )
    all_lines.extend(
        constraint_command.config_cmd(
            "Colocation Constraints:",
            lib.constraint_colocation.config,
            constraints_reports.colocation_plain,
            modifiers.get_subset("-f", "--full"),
        )
    )
    all_lines.extend(
        constraint_command.config_cmd(
            "Ticket Constraints:",
            lib.constraint_ticket.config,
            constraints_reports.ticket_plain,
            modifiers.get_subset("-f", "--full"),
        )
    )

    all_lines.append("")
    all_lines.extend(alert.alert_config_lines(lib))

    all_lines.append("")
    all_lines.append("Resources Defaults:")
    all_lines.extend(
        indent(
            nvset_dto_list_to_lines(
                lib.cib_options.resource_defaults_config(
                    evaluate_expired=False
                ).meta_attributes,
                nvset_label="Meta Attrs",
                with_ids=modifiers.get("--full"),
                text_if_empty="No defaults set",
            )
        )
    )
    all_lines.append("Operations Defaults:")
    all_lines.extend(
        indent(
            nvset_dto_list_to_lines(
                lib.cib_options.operation_defaults_config(
                    evaluate_expired=False
                ).meta_attributes,
                nvset_label="Meta Attrs",
                with_ids=modifiers.get("--full"),
                text_if_empty="No defaults set",
            )
        )
    )

    all_lines.append("")
    all_lines.append("Cluster Properties:")
    properties = utils.get_set_properties()
    all_lines.extend(
        indent(
            [
                "{0}: {1}".format(prop, val)
                for prop, val in sorted(properties.items())
            ],
            indent_step=1,
        )
    )
    all_lines.append("")
    all_lines.append("Tags:")
    tags = lib.tag.config([])
    if not tags:
        all_lines.append(" No tags defined")
    tag_lines = []
    for tag in tags:
        tag_lines.append(tag["tag_id"])
        tag_lines.extend(indent(tag["idref_list"]))
    all_lines.extend(indent(tag_lines, indent_step=1))
    return all_lines


def config_backup(lib, argv, modifiers):
    """
    Options:
      * --force - overwrite file if already exists
    """
    del lib
    modifiers.ensure_only_supported("--force")
    if len(argv) > 1:
        usage.config(["backup"])
        sys.exit(1)

    outfile_name = None
    if argv:
        outfile_name = argv[0]
        if not outfile_name.endswith(".tar.bz2"):
            outfile_name += ".tar.bz2"

    tar_data = config_backup_local()
    if outfile_name:
        ok, message = utils.write_file(
            outfile_name, tar_data, permissions=0o600, binary=True
        )
        if not ok:
            utils.err(message)
    else:
        # in python3 stdout accepts str so we need to use buffer
        sys.stdout.buffer.write(tar_data)


def config_backup_local():
    """
    Commandline options: no options
    """
    file_list = config_backup_path_list()
    tar_data = BytesIO()

    try:
        with tarfile.open(fileobj=tar_data, mode="w|bz2") as tarball:
            config_backup_add_version_to_tarball(tarball)
            for tar_path, path_info in file_list.items():
                if (
                    not os.path.exists(path_info["path"])
                    and not path_info["required"]
                ):
                    continue
                tarball.add(path_info["path"], tar_path)
    except (tarfile.TarError, EnvironmentError) as e:
        utils.err("unable to create tarball: %s" % e)

    tar = tar_data.getvalue()
    tar_data.close()
    return tar


def config_restore(lib, argv, modifiers):
    """
    Options:
      * --local - restore config only on local node
      * --request-timeout - timeout for HTTP requests, used only if --local was
        not defined or user is not root
    """
    del lib
    modifiers.ensure_only_supported("--local", "--request-timeout")
    if len(argv) > 1:
        usage.config(["restore"])
        sys.exit(1)

    infile_name = infile_obj = None
    if argv:
        infile_name = argv[0]
    if not infile_name:
        # in python3 stdin returns str so we need to use buffer
        infile_obj = BytesIO(sys.stdin.buffer.read())

    if os.getuid() == 0:
        if modifiers.get("--local"):
            config_restore_local(infile_name, infile_obj)
        else:
            config_restore_remote(infile_name, infile_obj)
    else:
        new_argv = ["config", "restore"]
        options = []
        new_stdin = None
        if modifiers.get("--local"):
            options.append("--local")
        if infile_name:
            new_argv.append(os.path.abspath(infile_name))
        else:
            new_stdin = infile_obj.read()
        err_msgs, exitcode, std_out, std_err = utils.call_local_pcsd(
            new_argv, options, new_stdin
        )
        if err_msgs:
            for msg in err_msgs:
                utils.err(msg, False)
            sys.exit(1)
        print(std_out)
        sys.stderr.write(std_err)
        sys.exit(exitcode)


def config_restore_remote(infile_name, infile_obj):
    """
    Commandline options:
      * --request-timeout - timeout for HTTP requests
    """
    extracted = {
        "version.txt": "",
        "corosync.conf": "",
    }
    try:
        with tarfile.open(infile_name, "r|*", infile_obj) as tarball:
            while True:
                # next(tarball) does not work in python2.6
                tar_member_info = tarball.next()
                if tar_member_info is None:
                    break
                if tar_member_info.name in extracted:
                    tar_member = tarball.extractfile(tar_member_info)
                    extracted[tar_member_info.name] = tar_member.read()
                    tar_member.close()
    except (tarfile.TarError, EnvironmentError) as e:
        utils.err("unable to read the tarball: %s" % e)

    config_backup_check_version(extracted["version.txt"])

    node_list, report_list = get_existing_nodes_names(
        utils.get_corosync_conf_facade(
            conf_text=extracted["corosync.conf"].decode("utf-8")
        )
    )
    if report_list:
        process_library_reports(report_list)
    if not node_list:
        utils.err("no nodes found in the tarball")

    err_msgs = []
    for node in node_list:
        try:
            retval, output = utils.checkStatus(node)
            if retval != 0:
                err_msgs.append(output)
                continue
            _status = json.loads(output)
            if (
                _status["corosync"]
                or _status["pacemaker"]
                or
                # not supported by older pcsd, do not fail if not present
                _status.get("pacemaker_remote", False)
            ):
                err_msgs.append(
                    "Cluster is currently running on node %s. You need to stop "
                    "the cluster in order to restore the configuration." % node
                )
                continue
        except (ValueError, NameError, LookupError):
            err_msgs.append("unable to determine status of the node %s" % node)
    if err_msgs:
        for msg in err_msgs:
            utils.err(msg, False)
        sys.exit(1)

    # Temporarily disable config files syncing thread in pcsd so it will not
    # rewrite restored files. 10 minutes should be enough time to restore.
    # If node returns HTTP 404 it does not support config syncing at all.
    for node in node_list:
        retval, output = utils.pauseConfigSyncing(node, 10 * 60)
        if not (retval == 0 or "(HTTP error: 404)" in output):
            utils.err(output)

    if infile_obj:
        infile_obj.seek(0)
        tarball_data = infile_obj.read()
    else:
        with open(infile_name, "rb") as tarball:
            tarball_data = tarball.read()

    error_list = []
    for node in node_list:
        retval, error = utils.restoreConfig(node, tarball_data)
        if retval != 0:
            error_list.append(error)
    if error_list:
        utils.err("unable to restore all nodes\n" + "\n".join(error_list))


def config_restore_local(infile_name, infile_obj):
    """
    Commandline options: no options
    """
    service_manager = utils.get_service_manager()
    if (
        service_manager.is_running("corosync")
        or service_manager.is_running("pacemaker")
        or service_manager.is_running("pacemaker_remote")
    ):
        utils.err(
            "Cluster is currently running on this node. You need to stop "
            "the cluster in order to restore the configuration."
        )

    file_list = config_backup_path_list(with_uid_gid=True)
    tarball_file_list = []
    version = None
    tmp_dir = None
    try:
        with tarfile.open(infile_name, "r|*", infile_obj) as tarball:
            while True:
                # next(tarball) does not work in python2.6
                tar_member_info = tarball.next()
                if tar_member_info is None:
                    break
                if tar_member_info.name == "version.txt":
                    version_data = tarball.extractfile(tar_member_info)
                    version = version_data.read()
                    version_data.close()
                    continue
                tarball_file_list.append(tar_member_info.name)

        required_file_list = [
            tar_path
            for tar_path, path_info in file_list.items()
            if path_info["required"]
        ]
        missing = set(required_file_list) - set(tarball_file_list)
        if missing:
            utils.err(
                "unable to restore the cluster, missing files in backup: %s"
                % ", ".join(missing)
            )

        config_backup_check_version(version)

        if infile_obj:
            infile_obj.seek(0)
        with tarfile.open(infile_name, "r|*", infile_obj) as tarball:
            while True:
                # next(tarball) does not work in python2.6
                tar_member_info = tarball.next()
                if tar_member_info is None:
                    break
                extract_info = None
                path = tar_member_info.name
                while path:
                    if path in file_list:
                        extract_info = file_list[path]
                        break
                    path = os.path.dirname(path)
                if not extract_info:
                    continue
                path_full = None
                if hasattr(extract_info.get("pre_store_call"), "__call__"):
                    extract_info["pre_store_call"]()
                if "rename" in extract_info and extract_info["rename"]:
                    if tmp_dir is None:
                        tmp_dir = tempfile.mkdtemp()
                    tarball.extractall(tmp_dir, [tar_member_info])
                    path_full = extract_info["path"]
                    shutil.move(
                        os.path.join(tmp_dir, tar_member_info.name), path_full
                    )
                else:
                    dir_path = os.path.dirname(extract_info["path"])
                    tarball.extractall(dir_path, [tar_member_info])
                    path_full = os.path.join(dir_path, tar_member_info.name)
                file_attrs = extract_info["attrs"]
                os.chmod(path_full, file_attrs["mode"])
                os.chown(path_full, file_attrs["uid"], file_attrs["gid"])
    except (tarfile.TarError, EnvironmentError, OSError) as e:
        utils.err("unable to restore the cluster: %s" % e)
    finally:
        if tmp_dir:
            shutil.rmtree(tmp_dir, ignore_errors=True)

    try:
        sig_path = os.path.join(settings.cib_dir, "cib.xml.sig")
        if os.path.exists(sig_path):
            os.remove(sig_path)
    except EnvironmentError as e:
        utils.err("unable to remove %s: %s" % (sig_path, e))


def config_backup_path_list(with_uid_gid=False):
    """
    Commandline options: no option
    NOTE: corosync.conf path may be altered using --corosync_conf
    """
    corosync_attrs = {
        "mtime": int(time.time()),
        "mode": 0o644,
        "uname": "root",
        "gname": "root",
        "uid": 0,
        "gid": 0,
    }
    corosync_authkey_attrs = dict(corosync_attrs)
    corosync_authkey_attrs["mode"] = 0o400
    cib_attrs = {
        "mtime": int(time.time()),
        "mode": 0o600,
        "uname": settings.pacemaker_uname,
        "gname": settings.pacemaker_gname,
    }
    if with_uid_gid:
        cib_attrs["uid"] = _get_uid(cib_attrs["uname"])
        cib_attrs["gid"] = _get_gid(cib_attrs["gname"])

    pcmk_authkey_attrs = dict(cib_attrs)
    pcmk_authkey_attrs["mode"] = 0o440
    file_list = {
        "cib.xml": {
            "path": os.path.join(settings.cib_dir, "cib.xml"),
            "required": True,
            "attrs": dict(cib_attrs),
        },
        "corosync_authkey": {
            "path": settings.corosync_authkey_file,
            "required": False,
            "attrs": corosync_authkey_attrs,
            "restore_procedure": None,
            "rename": True,
        },
        "pacemaker_authkey": {
            "path": settings.pacemaker_authkey_file,
            "required": False,
            "attrs": pcmk_authkey_attrs,
            "restore_procedure": None,
            "rename": True,
            "pre_store_call": _ensure_etc_pacemaker_exists,
        },
        "corosync.conf": {
            "path": settings.corosync_conf_file,
            "required": True,
            "attrs": dict(corosync_attrs),
        },
        "uidgid.d": {
            "path": settings.corosync_uidgid_dir,
            "required": False,
            "attrs": dict(corosync_attrs),
        },
        "pcs_settings.conf": {
            "path": settings.pcsd_settings_conf_location,
            "required": False,
            "attrs": {
                "mtime": int(time.time()),
                "mode": 0o644,
                "uname": "root",
                "gname": "root",
                "uid": 0,
                "gid": 0,
            },
        },
    }
    return file_list


def _get_uid(user_name):
    """
    Commandline options: no options
    """
    try:
        return pwd.getpwnam(user_name).pw_uid
    except KeyError:
        return utils.err(
            "Unable to determine uid of user '{0}'".format(user_name)
        )


def _get_gid(group_name):
    """
    Commandline options: no options
    """
    try:
        return grp.getgrnam(group_name).gr_gid
    except KeyError:
        return utils.err(
            "Unable to determine gid of group '{0}'".format(group_name)
        )


def _ensure_etc_pacemaker_exists():
    """
    Commandline options: no options
    """
    dir_name = os.path.dirname(settings.pacemaker_authkey_file)
    if not os.path.exists(dir_name):
        os.mkdir(dir_name)
        os.chmod(dir_name, 0o750)
        os.chown(
            dir_name,
            _get_uid(settings.pacemaker_uname),
            _get_gid(settings.pacemaker_gname),
        )


def config_backup_check_version(version):
    """
    Commandline options: no options
    """
    try:
        version_number = int(version)
        supported_version = config_backup_version()
        if version_number > supported_version:
            utils.err(
                "Unsupported version of the backup, "
                "supported version is %d, backup version is %d"
                % (supported_version, version_number)
            )
        if version_number < supported_version:
            print(
                "Warning: restoring from the backup version %d, "
                "current supported version is %s"
                % (version_number, supported_version)
            )
    except TypeError:
        utils.err("Cannot determine version of the backup")


def config_backup_add_version_to_tarball(tarball, version=None):
    """
    Commandline options: no options
    """
    ver = version if version is not None else str(config_backup_version())
    return utils.tar_add_file_data(tarball, ver.encode("utf-8"), "version.txt")


def config_backup_version():
    """
    Commandline options: no options
    """
    return 1


def config_checkpoint_list(lib, argv, modifiers):
    """
    Options: no options
    """
    del lib
    modifiers.ensure_only_supported()
    if argv:
        raise CmdLineInputError()
    try:
        file_list = os.listdir(settings.cib_dir)
    except OSError as e:
        utils.err("unable to list checkpoints: %s" % e)
    cib_list = []
    cib_name_re = re.compile(r"^cib-(\d+)\.raw$")
    for filename in file_list:
        match = cib_name_re.match(filename)
        if not match:
            continue
        file_path = os.path.join(settings.cib_dir, filename)
        try:
            if os.path.isfile(file_path):
                cib_list.append(
                    (float(os.path.getmtime(file_path)), match.group(1))
                )
        except OSError:
            pass
    cib_list.sort()
    if not cib_list:
        print("No checkpoints available")
        return
    for cib_info in cib_list:
        print(
            "checkpoint %s: date %s"
            % (cib_info[1], datetime.datetime.fromtimestamp(round(cib_info[0])))
        )


def _checkpoint_to_lines(lib, checkpoint_number):
    # backup current settings
    orig_usefile = utils.usefile
    orig_filename = utils.filename
    orig_middleware = lib.middleware_factory
    # configure old code to read the CIB from a file
    utils.usefile = True
    utils.filename = os.path.join(
        settings.cib_dir, "cib-%s.raw" % checkpoint_number
    )
    # configure new code to read the CIB from a file
    lib.middleware_factory = orig_middleware._replace(
        cib=middleware.cib(utils.filename, utils.touch_cib_file)
    )
    # export the CIB to text
    result = False, []
    if os.path.isfile(utils.filename):
        result = True, _config_show_cib_lines(lib)
    # restore original settings
    utils.usefile = orig_usefile
    utils.filename = orig_filename
    lib.middleware_factory = orig_middleware
    return result


def config_checkpoint_view(lib, argv, modifiers):
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if len(argv) != 1:
        usage.config(["checkpoint view"])
        sys.exit(1)

    loaded, lines = _checkpoint_to_lines(lib, argv[0])
    if not loaded:
        utils.err("unable to read the checkpoint")
    print("\n".join(lines))


def config_checkpoint_diff(lib, argv, modifiers):
    """
    Commandline options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    if len(argv) != 2:
        usage.config(["checkpoint diff"])
        sys.exit(1)

    if argv[0] == argv[1]:
        utils.err("cannot diff a checkpoint against itself")

    errors = []
    checkpoints_lines = []
    for checkpoint in argv:
        if checkpoint == "live":
            lines = _config_show_cib_lines(lib)
            if not lines:
                errors.append("unable to read live configuration")
            else:
                checkpoints_lines.append(lines)
        else:
            loaded, lines = _checkpoint_to_lines(lib, checkpoint)
            if not loaded:
                errors.append(
                    "unable to read checkpoint '{0}'".format(checkpoint)
                )
            else:
                checkpoints_lines.append(lines)

    if errors:
        utils.err("\n".join(errors))

    print(
        "Differences between {0} (-) and {1} (+):".format(
            *[
                "live configuration"
                if label == "live"
                else f"checkpoint {label}"
                for label in argv
            ]
        )
    )
    print(
        "\n".join(
            [
                line.rstrip()
                for line in difflib.Differ().compare(
                    checkpoints_lines[0], checkpoints_lines[1]
                )
            ]
        )
    )


def config_checkpoint_restore(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file, a checkpoint will be restored into a specified file
    """
    # pylint: disable=broad-except
    del lib
    modifiers.ensure_only_supported("-f")
    if len(argv) != 1:
        usage.config(["checkpoint restore"])
        sys.exit(1)

    cib_path = os.path.join(settings.cib_dir, "cib-%s.raw" % argv[0])
    try:
        snapshot_dom = parse(cib_path)
    except Exception as e:
        utils.err("unable to read the checkpoint: %s" % e)
    utils.replace_cib_configuration(snapshot_dom)


# TODO remove, deprecated command
def config_import_cman(lib, argv, modifiers):
    """
    Options:
      * --force - skip checks, overwrite files
      * --interactive - interactive issue resolving
      * --request-timeout - effective only when output is not specified
    """
    # pylint: disable=no-member
    del lib
    warn("This command is deprecated and will be removed.")
    modifiers.ensure_only_supported(
        "--force",
        "interactive",
        "--request-timeout",
    )
    if no_clufter:
        utils.err(
            "Unable to perform a CMAN cluster conversion due to missing "
            "python-clufter package"
        )
    clufter_supports_corosync3 = hasattr(clufter.facts, "cluster_pcs_camelback")

    # prepare convertor options
    cluster_conf = settings.cluster_conf_file
    dry_run_output = None
    output_format = "corosync.conf"
    dist = None
    invalid_args = False
    for arg in argv:
        if "=" in arg:
            name, value = arg.split("=", 1)
            if name == "input":
                cluster_conf = value
            elif name == "output":
                dry_run_output = value
            elif name == "output-format":
                if value in (
                    "corosync.conf",
                    "pcs-commands",
                    "pcs-commands-verbose",
                ):
                    output_format = value
                else:
                    invalid_args = True
            elif name == "dist":
                dist = value
            else:
                invalid_args = True
        else:
            invalid_args = True
    if output_format not in ("pcs-commands", "pcs-commands-verbose") and (
        dry_run_output and not dry_run_output.endswith(".tar.bz2")
    ):
        dry_run_output += ".tar.bz2"
    if invalid_args or not dry_run_output:
        usage.config(["import-cman"])
        sys.exit(1)
    debug = modifiers.get("--debug")
    force = modifiers.get("--force")
    interactive = modifiers.get("--interactive")

    if dist is not None:
        if not clufter_supports_corosync3:
            utils.err(
                "Unable to perform a CMAN cluster conversion due to clufter "
                "not supporting Corosync 3. Please, upgrade clufter packages."
            )
        if not clufter.facts.cluster_pcs_camelback("linux", dist.split(",")):
            utils.err("dist does not match output-format")
    elif output_format == "corosync.conf":
        dist = _get_linux_dist()
    else:
        # for output-format=pcs-command[-verbose]
        dist = _get_linux_dist()

    clufter_args = {
        "input": str(cluster_conf),
        "cib": {"passin": "bytestring"},
        "nocheck": force,
        "batch": True,
        "sys": "linux",
        "dist": dist,
    }
    if interactive:
        if "EDITOR" not in os.environ:
            utils.err("$EDITOR environment variable is not set")
        clufter_args["batch"] = False
        clufter_args["editor"] = os.environ["EDITOR"]
    if debug:
        logging.getLogger("clufter").setLevel(logging.DEBUG)
    if output_format == "corosync.conf":
        clufter_args["coro"] = {"passin": "struct"}
        cmd_name = "ccs2pcs-camelback"
    elif output_format in ("pcs-commands", "pcs-commands-verbose"):
        clufter_args["output"] = {"passin": "bytestring"}
        clufter_args["start_wait"] = "60"
        clufter_args["tmp_cib"] = "tmp-cib.xml"
        clufter_args["force"] = force
        clufter_args["text_width"] = "80"
        clufter_args["silent"] = True
        clufter_args["noguidance"] = True
        if output_format == "pcs-commands-verbose":
            clufter_args["text_width"] = "-1"
            clufter_args["silent"] = False
            clufter_args["noguidance"] = False
        if clufter.facts.cluster_pcs_flatiron("linux", dist.split(",")):
            cmd_name = "ccs2pcscmd-flatiron"
        elif clufter.facts.cluster_pcs_needle("linux", dist.split(",")):
            cmd_name = "ccs2pcscmd-needle"
        elif clufter_supports_corosync3 and clufter.facts.cluster_pcs_camelback(
            "linux", dist.split(",")
        ):
            cmd_name = "ccs2pcscmd-camelback"
        else:
            utils.err(
                "unrecognized dist, try something recognized"
                + " (e. g. rhel,6.8 or redhat,7.3 or debian,7 or ubuntu,trusty)"
            )
    clufter_args_obj = type(str("ClufterOptions"), (object,), clufter_args)

    # run convertor
    run_clufter(
        cmd_name,
        clufter_args_obj,
        debug,
        force,
        "Error: unable to import cluster configuration",
    )

    # save commands
    if output_format in ("pcs-commands", "pcs-commands-verbose"):
        ok, message = utils.write_file(
            dry_run_output, clufter_args_obj.output["passout"].decode()
        )
        if not ok:
            utils.err(message)
        return

    # put new config files into tarball
    file_list = config_backup_path_list()
    for file_item in file_list.values():
        file_item["attrs"]["uname"] = "root"
        file_item["attrs"]["gname"] = "root"
        file_item["attrs"]["uid"] = 0
        file_item["attrs"]["gid"] = 0
        file_item["attrs"]["mode"] = 0o600
    tar_data = BytesIO()
    try:
        with tarfile.open(fileobj=tar_data, mode="w|bz2") as tarball:
            config_backup_add_version_to_tarball(tarball)
            utils.tar_add_file_data(
                tarball,
                clufter_args_obj.cib["passout"],
                "cib.xml",
                **file_list["cib.xml"]["attrs"],
            )
            # put uidgid into separate files
            fmt_simpleconfig = clufter.format_manager.FormatManager.init_lookup(
                "simpleconfig"
            ).plugins["simpleconfig"]
            corosync_struct = []
            uidgid_list = []
            for section in clufter_args_obj.coro["passout"][2]:
                if section[0] == "uidgid":
                    uidgid_list.append(section[1])
                else:
                    corosync_struct.append(section)
            corosync_conf_data = fmt_simpleconfig(
                "struct", ("corosync", (), corosync_struct)
            )("bytestring")
            utils.tar_add_file_data(
                tarball,
                corosync_conf_data,
                "corosync.conf",
                **file_list["corosync.conf"]["attrs"],
            )
            for uidgid in uidgid_list:
                uid = ""
                gid = ""
                for item in uidgid:
                    if item[0] == "uid":
                        uid = item[1]
                    if item[0] == "gid":
                        gid = item[1]
                filename = utils.get_uid_gid_file_name(uid, gid)
                uidgid_data = fmt_simpleconfig(
                    "struct", ("corosync", (), [("uidgid", uidgid, None)])
                )("bytestring")
                utils.tar_add_file_data(
                    tarball,
                    uidgid_data,
                    "uidgid.d/" + filename,
                    **file_list["uidgid.d"]["attrs"],
                )
    except (tarfile.TarError, EnvironmentError) as e:
        utils.err("unable to create tarball: %s" % e)
    tar_data.seek(0)

    # save tarball / remote restore
    if dry_run_output:
        ok, message = utils.write_file(
            dry_run_output, tar_data.read(), permissions=0o600, binary=True
        )
        if not ok:
            utils.err(message)
    else:
        config_restore_remote(None, tar_data)
    tar_data.close()


def _get_linux_dist():
    if no_distro_package:
        # For Python 3.8+, python3-distro is a required dependency and we
        # should never get here. Pylint, of course, cannot know that.
        # pylint: disable=deprecated-method
        # pylint: disable=no-member
        # pylint: disable=used-before-assignment
        distribution = platform.linux_distribution(full_distribution_name=False)
    else:
        distribution = distro.linux_distribution(full_distribution_name=False)
    return ",".join(distribution)


# TODO remove, deprecated command
def config_export_pcs_commands(lib, argv, modifiers, verbose=False):
    """
    Options:
      * --force - skip checks, overwrite files
      * --interactive - interactive issue resolving
      * -f - CIB file
      * --corosync_conf
    """
    del lib
    warn("This command is deprecated and will be removed.")
    modifiers.ensure_only_supported(
        "--force", "--interactive", "-f", "--corosync_conf"
    )
    if no_clufter:
        utils.err(
            "Unable to perform export due to missing python-clufter package"
        )

    # parse options
    debug = modifiers.get("--debug")
    force = modifiers.get("--force")
    interactive = modifiers.get("--interactive")
    invalid_args = False
    output_file = None
    dist = None
    for arg in argv:
        if "=" in arg:
            name, value = arg.split("=", 1)
            if name == "output":
                output_file = value
            elif name == "dist":
                dist = value
            else:
                invalid_args = True
        else:
            invalid_args = True
    # check options
    if invalid_args:
        usage.config(["export pcs-commands"])
        sys.exit(1)
    # complete optional options
    if dist is None:
        dist = _get_linux_dist()

    # prepare convertor options
    clufter_args = {
        "nocheck": force,
        "batch": True,
        "sys": "linux",
        "dist": dist,
        "coro": settings.corosync_conf_file,
        "start_wait": "60",
        "tmp_cib": "tmp-cib.xml",
        "force": force,
        "text_width": "80",
        "silent": True,
        "noguidance": True,
    }
    if output_file:
        clufter_args["output"] = {"passin": "bytestring"}
    else:
        clufter_args["output"] = "-"
    if interactive:
        if "EDITOR" not in os.environ:
            utils.err("$EDITOR environment variable is not set")
        clufter_args["batch"] = False
        clufter_args["editor"] = os.environ["EDITOR"]
    if debug:
        logging.getLogger("clufter").setLevel(logging.DEBUG)
    if utils.usefile:
        clufter_args["cib"] = os.path.abspath(utils.filename)
    else:
        clufter_args["cib"] = ("bytestring", utils.get_cib())
    if verbose:
        clufter_args["text_width"] = "-1"
        clufter_args["silent"] = False
        clufter_args["noguidance"] = False
    clufter_args_obj = type(str("ClufterOptions"), (object,), clufter_args)
    cmd_name = "pcs2pcscmd-camelback"

    # run convertor
    run_clufter(
        cmd_name,
        clufter_args_obj,
        debug,
        force,
        "Error: unable to export cluster configuration",
    )

    # save commands if not printed to stdout by clufter
    if output_file:
        # pylint: disable=no-member
        ok, message = utils.write_file(
            output_file, clufter_args_obj.output["passout"].decode()
        )
        if not ok:
            utils.err(message)


# TODO remove, deprecated
def run_clufter(cmd_name, cmd_args, debug, force, err_prefix):
    """
    Commandline options: no options used but messages which include --force,
      --debug and --interactive are generated
    """
    # pylint: disable=broad-except
    try:
        result = None
        cmd_manager = clufter.command_manager.CommandManager.init_lookup(
            cmd_name
        )
        result = cmd_manager.commands[cmd_name](cmd_args)
        error_message = ""
    except Exception as e:
        error_message = str(e)
    if error_message or result != 0:
        hints = []
        hints.append("--interactive to solve the issues manually")
        if not debug:
            hints.append("--debug to get more information")
        if not force:
            hints.append("--force to override")
        hints_string = "\nTry using %s." % ", ".join(hints) if hints else ""
        sys.stderr.write(
            err_prefix
            + (": %s" % error_message if error_message else "")
            + hints_string
            + "\n"
        )
        sys.exit(1 if result is None else result)



from typing import (
    Any,
    Mapping,
    Sequence,
)

from dataclasses import dataclass

from pcs.common.interface.dto import DataTransferObject
from pcs.common.types import ResourceRelationType


@dataclass(frozen=True)
class RelationEntityDto(DataTransferObject):
    id: str  # pylint: disable=invalid-name
    type: ResourceRelationType
    members: Sequence[str]
    metadata: Mapping[str, Any]


@dataclass(frozen=True)
class ResourceRelationDto(DataTransferObject):
    relation_entity: RelationEntityDto
    members: Sequence["ResourceRelationDto"]
    is_leaf: bool



from unittest import (
    TestCase,
    mock,
)

from pcs import host
from pcs.cli.common.errors import CmdLineInputError

from pcs_test.tools.misc import dict_to_modifiers


class HostAuth(TestCase):
    host_names = {
        "node-01": {
            "hostname": "node-01.example.com",
            "ipv4": "192.168.122.11",
            "ipv6": "[2620:52:0:25a4:1800:ff:fe00:1]",
        },
        "node-02": {
            "hostname": "node-02.example.com",
            "ipv4": "192.168.122.12",
            "ipv6": "[2620:52:0:25a4:1800:ff:fe00:2]",
        },
    }

    def setUp(self):
        self.lib = mock.Mock()
        self.patch_get_user_and_pass = mock.patch(
            "pcs.utils.get_user_and_pass",
            return_value=("hacluster", "password"),
        )
        self.patch_auth_hosts = mock.patch("pcs.utils.auth_hosts")

    @staticmethod
    def _fixture_args(name_addr_port_tuple_list):
        arg_list = []
        for name, addr, port in name_addr_port_tuple_list:
            port = ":{}".format(port) if port is not None else ""
            arg_list.extend(
                [
                    name,
                    "addr={}{}".format(addr, port),
                ]
            )
        return arg_list

    def _assert_invalid_port(self, name_addr_port_tuple_list):
        arg_list = self._fixture_args(name_addr_port_tuple_list)
        mock_auth_hosts = self.patch_auth_hosts.start()
        with self.assertRaises(CmdLineInputError) as cm:
            host.auth_cmd(self.lib, arg_list, dict_to_modifiers({}))
        _, addr, port = name_addr_port_tuple_list[-1]
        self.assertEqual(
            (
                "Invalid port number in address '{addr}:{port}', use 1..65535"
            ).format(addr=addr, port=port),
            cm.exception.message,
        )
        mock_auth_hosts.assert_not_called()
        self.patch_auth_hosts.stop()

    def _assert_valid_port(self, name_addr_port_tuple_list):
        arg_list = self._fixture_args(name_addr_port_tuple_list)
        mock_get_user_and_pass = self.patch_get_user_and_pass.start()
        mock_auth_hosts = self.patch_auth_hosts.start()
        host.auth_cmd(self.lib, arg_list, dict_to_modifiers({}))
        mock_get_user_and_pass.assert_called_once_with()
        mock_auth_hosts.assert_called_once_with(
            {
                name: {
                    "dest_list": [
                        dict(
                            addr=addr
                            if addr.count(":") <= 1
                            else addr.strip("[]"),
                            port=port,
                        )
                    ],
                    "username": "hacluster",
                    "password": "password",
                }
                for name, addr, port in name_addr_port_tuple_list
            }
        )
        self.patch_get_user_and_pass.stop()
        self.patch_auth_hosts.stop()

    def run_port_subtests(self, assert_function, port_list):
        for addr_type in ["hostname", "ipv4", "ipv6"]:
            name_list = sorted(self.host_names.keys())[0 : len(port_list)]
            addr_list = [self.host_names[name][addr_type] for name in name_list]

            with self.subTest(addr_type=addr_type):
                assert_function(list(zip(name_list, addr_list, port_list)))

    @mock.patch("pcs.utils.auth_hosts")
    def test_no_args(self, mock_auth_hosts):
        with self.assertRaises(CmdLineInputError) as cm:
            host.auth_cmd(self.lib, [], dict_to_modifiers({}))
        self.assertEqual("No host specified", cm.exception.message)
        mock_auth_hosts.assert_not_called()

    def test_invalid_port_notanumber(self):
        self.run_port_subtests(self._assert_invalid_port, ["notanumber"])

    def test_invalid_port_lower_bound(self):
        self.run_port_subtests(self._assert_invalid_port, [0])

    def test_invalid_port_higher_bound(self):
        self.run_port_subtests(self._assert_invalid_port, [65536])

    def test_invalid_port_notanumber_multinode(self):
        self.run_port_subtests(
            self._assert_invalid_port,
            [1, "notanumber"],
        )

    def test_invalid_port_lower_bound_multinode(self):
        self.run_port_subtests(self._assert_invalid_port, [3000, 0])

    def test_invalid_port_higher_bound_hostname_multinode(self):
        self.run_port_subtests(self._assert_invalid_port, [65535, 65536])

    def test_valid_port(self):
        self.run_port_subtests(self._assert_valid_port, [3000])

    def test_valid_port_multinode(self):
        self.run_port_subtests(self._assert_valid_port, [3000, 5000])



from typing import Sequence

from dataclasses import dataclass

from pcs.common.interface.dto import DataTransferObject

from .bundle import CibResourceBundleDto
from .clone import CibResourceCloneDto
from .group import CibResourceGroupDto
from .primitive import CibResourcePrimitiveDto


@dataclass(frozen=True)
class ListCibResourcesDto(DataTransferObject):
    primitives: Sequence[CibResourcePrimitiveDto]
    clones: Sequence[CibResourceCloneDto]
    groups: Sequence[CibResourceGroupDto]
    bundles: Sequence[CibResourceBundleDto]



from typing import (
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Union,
    cast,
)

from lxml import etree
from lxml.etree import (
    _Element,
    _ElementTree,
)

from pcs.common.types import StringCollection


def get_root(tree: Union[_Element, _ElementTree]) -> _Element:
    # ElementTree has getroot, Element has getroottree
    if isinstance(tree, _ElementTree):
        return tree.getroot()
    # getroot() turns _ElementTree to _Element
    return tree.getroottree().getroot()


def find_parent(
    element: _Element, tag_names: StringCollection
) -> Optional[_Element]:
    """
    Return the closest parent with specified tag name of an element or None

    element -- the element whose parent we want to find
    tag_names -- allowed tag names of a parent we are looking for
    """
    candidate: Optional[_Element] = element
    while True:
        if candidate is None or candidate.tag in tag_names:
            return candidate
        candidate = candidate.getparent()


def get_sub_element(
    element: _Element,
    sub_element_tag: str,
    new_id: Optional[str] = None,
    new_index: Optional[int] = None,
    append_if_missing: bool = True,
) -> _Element:
    """
    Returns the FIRST sub-element sub_element_tag of element. It will create new
    element if such doesn't exist yet.

    element -- parent element
    sub_element_tag -- tag of the wanted new element
    new_id -- id of the new element, None means no id will be set
    new_index -- where the new element will be added, None means at the end
    append_if_missing -- if the searched element does not exist, append it to
        the parent element
    """
    sub_element_list = cast(
        List[_Element],
        element.xpath("./*[local-name()=$tag_name]", tag_name=sub_element_tag),
    )
    if not sub_element_list:
        sub_element = etree.Element(sub_element_tag)
        if new_id:
            sub_element.set("id", new_id)
        if append_if_missing:
            if new_index is None:
                element.append(sub_element)
            else:
                element.insert(new_index, sub_element)
        return sub_element
    return sub_element_list[0]


def export_attributes(
    element: _Element, with_id: bool = True
) -> Dict[str, str]:
    result = {str(key): str(value) for key, value in element.attrib.items()}
    if not with_id:
        result.pop("id", None)
    return result


def update_attribute_remove_empty(
    element: _Element, name: str, value: str
) -> None:
    """
    Set an attribute's value or remove the attribute if the value is empty

    element -- element to be updated
    name -- attribute name
    value -- attribute value
    """
    if not value:
        if name in element.attrib:
            del element.attrib[name]
        return
    element.set(name, value)


def update_attributes_remove_empty(
    element: _Element, attributes: Mapping[str, str]
) -> None:
    """
    Set an attributes' values or remove an attribute if its new value is empty

    element -- element to be updated
    attributes -- new attributes' values
    """
    for name, value in attributes.items():
        update_attribute_remove_empty(element, name, value)


def etree_to_str(tree: _Element) -> str:
    """
    Export a lxml tree to a string

    tree - the tree to be exported
    """
    # etree returns string in bytes: b'xml'
    # python 3 removed .encode() from byte strings
    # run(...) calls subprocess.Popen.communicate which calls encode...
    # so there is bytes to str conversion
    raw = etree.tostring(tree)
    return raw.decode() if isinstance(raw, bytes) else raw


def is_element_useful(
    element: _Element, attribs_important: bool = True
) -> bool:
    """
    Is an element worth keeping?

    Some elements serve as a container for sub-elements. When all sub-elements
    are removed it is time to consider if such element is still meaningful.

    Some of these elements can be meaningful standalone when they contain
    attributes (e.g. "network" or "storage" in "bundle"). Some of these
    elements are not meaningful without sub-elements even if they have
    attributes (e.g. rsc_ticket - after last sub-element 'resource_set' removal
    there can be attributes but the element is pointless - more details at the
    appropriate place of use). By default, an element is meaningful when it
    contains attributes (except id) even if it has no sub-elements. This can be
    switched by attribs_important parameter.

    element -- element to analyze
    attribs_important -- if True, the element is useful if it contains
        attributes even if it has no sub-elements
    """
    return len(element) > 0 or (
        attribs_important
        and bool(element.attrib)
        and element.attrib.keys() != ["id"]
    )


def append_when_useful(
    parent: _Element,
    element: _Element,
    attribs_important: bool = True,
    index: Optional[int] = None,
) -> _Element:
    """
    Append an element to a parent if the element is useful (see
        is_element_useful for details)

    parent -- where to append the element
    element -- the element to append
    attribs_important -- if True, append even if the element has no children if
        it has attributes
    index -- position to append the element, None means at the end
    """
    if element.getparent() == parent:
        return element
    if is_element_useful(element, attribs_important):
        if index is None:
            parent.append(element)
        else:
            parent.insert(index, element)
    return element


def remove_when_pointless(
    element: _Element,
    attribs_important: bool = True,
) -> None:
    """
    Remove an element when it is not worth keeping (see is_element_useful for
        details).

    element -- element to remove
    attribs_important -- if True, do not delete the element if it contains
        attributes
    """
    if not is_element_useful(element, attribs_important):
        remove_one_element(element)


def reset_element(
    element: _Element, keep_attrs: Optional[StringCollection] = None
) -> None:
    """
    Remove all subelements and all attributes (except mentioned in keep_attrs)
    of given element.

    element -- element to reset
    keep_attrs -- names of attributes thas should be kept
    """
    keep_attrs = keep_attrs or []
    for child in list(element):
        element.remove(child)
    for key in element.attrib.keys():
        if key not in keep_attrs:
            del element.attrib[key]


def move_elements(
    to_move_list: Iterable[_Element],
    adjacent_el: _Element,
    put_after_adjacent: bool = False,
) -> None:
    """
    Move elements inside or into an element after or before specified element
    in the element.

    to_move_list -- elements to be moved
    adjacent_el -- the element next to which the moved elements will be put
    put_after_adjacent -- put elements after (True) or before (False) the
        adjacent element
    """
    for el in to_move_list:
        if put_after_adjacent:
            adjacent_el.addnext(el)
            adjacent_el = el
        else:
            adjacent_el.addprevious(el)


def remove_one_element(element: _Element) -> None:
    """
    Remove single specified element.

    element -- element to be removed
    """
    parent = element.getparent()
    if parent is not None:
        parent.remove(element)



from typing import (
    Dict,
    cast,
)

from lxml.etree import _Element

from pcs.common.str_tools import (
    format_name_value_list,
    quote,
)
from pcs.lib.xml_tools import export_attributes


class RuleToStr:
    """
    Export a rule XML element to a string which creates the same element
    """

    def __init__(self) -> None:
        # The cache prevents evaluating subtrees repeatedly.
        self._cache: Dict[str, str] = {}

    def get_str(self, rule_part_element: _Element) -> str:
        """
        Export a rule XML element to a string which creates the same element

        rule_part_element -- a rule or rule expression element to be exported
        """
        element_id = str(rule_part_element.get("id", ""))
        if element_id not in self._cache:
            self._cache[element_id] = self._export_element(rule_part_element)
        return self._cache[element_id]

    def _export_element(self, rule_el: _Element) -> str:
        return self._tag_to_export[str(rule_el.tag)](self, rule_el)

    @staticmethod
    def _attrs_to_str(el: _Element) -> str:
        return " ".join(
            format_name_value_list(
                sorted(export_attributes(el, with_id=False).items())
            )
        )

    def _rule_to_str(self, rule_el: _Element) -> str:
        # "and" is a documented pacemaker default
        # https://clusterlabs.org/pacemaker/doc/en-US/Pacemaker/2.0/html-single/Pacemaker_Explained/index.html#_rule_properties
        boolean_op = str(rule_el.get("boolean-op", "and"))
        string_parts = []
        for child in cast(_Element, rule_el.xpath(self._xpath_for_export)):
            if child.tag == "rule":
                string_parts.append(f"({self.get_str(child)})")
            else:
                string_parts.append(self.get_str(child))
        return f" {boolean_op} ".join(string_parts)

    def _simple_expr_to_str(self, expr_el: _Element) -> str:
        # pylint - all *_to_str methods must have the same interface
        # pylint: disable=no-self-use
        string_parts = []
        if "value" in expr_el.attrib:
            # "attribute" and "operation" are defined as mandatory in CIB schema
            string_parts.extend(
                [
                    str(expr_el.get("attribute", "")),
                    str(expr_el.get("operation", "")),
                ]
            )
            if "type" in expr_el.attrib:
                string_parts.append(str(expr_el.get("type", "")))
            string_parts.append(quote(str(expr_el.get("value", "")), " "))
        else:
            # "attribute" and "operation" are defined as mandatory in CIB schema
            string_parts.extend(
                [
                    str(expr_el.get("operation", "")),
                    str(expr_el.get("attribute", "")),
                ]
            )
        return " ".join(string_parts)

    def _date_expr_to_str(self, expr_el: _Element) -> str:
        date_spec = expr_el.find("./date_spec")
        duration = expr_el.find("./duration")

        string_parts = []
        # "operation" is defined as mandatory in CIB schema
        operation = expr_el.get("operation", "")
        if operation == "date_spec":
            string_parts.append("date-spec")
            if date_spec is not None:
                string_parts.append(self._attrs_to_str(date_spec))
        elif operation == "in_range":
            string_parts.extend(["date", "in_range"])
            # CIB schema allows "start" + "duration" or optional "start" + "end"
            if "start" in expr_el.attrib:
                string_parts.append(str(expr_el.get("start", "")))
            string_parts.append("to")
            if "end" in expr_el.attrib:
                string_parts.append(str(expr_el.get("end", "")))
            if duration is not None:
                string_parts.append("duration")
                string_parts.append(self._attrs_to_str(duration))
        else:
            # CIB schema allows operation=="gt" + "start" or
            # operation=="lt" + "end"
            string_parts.extend(["date", str(expr_el.get("operation", ""))])
            if "start" in expr_el.attrib:
                string_parts.append(str(expr_el.get("start", "")))
            if "end" in expr_el.attrib:
                string_parts.append(str(expr_el.get("end", "")))
        return " ".join(string_parts)

    def _op_expr_to_str(self, expr_el: _Element) -> str:
        # pylint - all *_to_str methods must have the same interface
        # pylint: disable=no-self-use
        string_parts = ["op", str(expr_el.get("name", ""))]
        if "interval" in expr_el.attrib:
            string_parts.append(
                "interval={0}".format(expr_el.get("interval", ""))
            )
        return " ".join(string_parts)

    def _rsc_expr_to_str(self, expr_el: _Element) -> str:
        # pylint - all *_to_str methods must have the same interface
        # pylint: disable=no-self-use
        return "resource " + ":".join(
            [
                str(expr_el.get(attr, ""))
                for attr in ["class", "provider", "type"]
            ]
        )

    _tag_to_export = {
        "rule": _rule_to_str,
        "expression": _simple_expr_to_str,
        "date_expression": _date_expr_to_str,
        "op_expression": _op_expr_to_str,
        "rsc_expression": _rsc_expr_to_str,
    }

    _xpath_for_export = "./*[{export_tags}]".format(
        export_tags=" or ".join(f"self::{tag}" for tag in _tag_to_export)
    )



import datetime


class Writer:
    # pylint: disable=invalid-name
    def __init__(
        self,
        stream,
        format_,
        descriptions,
        traceback_highlight=False,
        fast_info=False,
    ):
        self.stream = stream
        self.format = format_
        self.descriptions = descriptions
        self.traceback_highlight = traceback_highlight
        self.fast_info = fast_info

    def addSuccess(self, test):
        pass

    def addError(self, test, err, traceback):
        pass

    def addFailure(self, test, err, traceback):
        pass

    def addSkip(self, test, reason):
        pass

    def startTest(self, test):
        pass

    def addExpectedFailure(self, test, err):
        pass

    def addUnexpectedSuccess(self, test):
        pass

    def show_fast_info(self, traceback):
        if self.fast_info:
            self.stream.writeln()
            self.stream.writeln(
                self.format.traceback(traceback)
                if self.traceback_highlight
                else traceback
            )


class DotWriter(Writer):
    def addSuccess(self, test):
        self.stream.write(self.format.green("."))
        self.stream.flush()

    def addError(self, test, err, traceback):
        self.stream.write(self.format.red("E"))
        self.stream.flush()
        self.show_fast_info(traceback)

    def addFailure(self, test, err, traceback):
        self.stream.write(self.format.red("F"))
        self.stream.flush()
        self.show_fast_info(traceback)

    def addSkip(self, test, reason):
        self.stream.write(self.format.blue("s"))
        self.stream.flush()

    def addExpectedFailure(self, test, err):
        self.stream.write(self.format.blue("x"))
        self.stream.flush()

    def addUnexpectedSuccess(self, test):
        self.stream.write(self.format.red("u"))
        self.stream.flush()


class StandardVerboseWriter(Writer):
    def addSuccess(self, test):
        self.stream.writeln(self.format.green("OK"))

    def addError(self, test, err, traceback):
        self.stream.writeln(self.format.red("ERROR"))
        self.show_fast_info(traceback)

    def addFailure(self, test, err, traceback):
        self.stream.writeln(self.format.red("FAIL"))
        self.show_fast_info(traceback)

    def addSkip(self, test, reason):
        self.stream.writeln(self.format.blue("skipped {0!r}".format(reason)))

    def startTest(self, test):
        self.stream.write(self.format.description(test, self.descriptions))
        self.stream.write(" ... ")
        self.stream.flush()

    def addExpectedFailure(self, test, err):
        self.stream.writeln(self.format.blue("expected failure"))

    def addUnexpectedSuccess(self, test):
        self.stream.writeln(self.format.red("unexpected success"))


class TimeWriter(StandardVerboseWriter):
    def __init__(
        self,
        stream,
        format_,
        descriptions,
        traceback_highlight=False,
        fast_info=False,
    ):
        super().__init__(
            stream,
            format_,
            descriptions,
            traceback_highlight=traceback_highlight,
            fast_info=fast_info,
        )
        self._start_time_map = {}

    def _print_time(self, test):
        end = datetime.datetime.now()
        testname = self.format.test_name(test)
        delta = (end - self._start_time_map[testname]).total_seconds()
        self.stream.writeln(f"{delta:11.6f}s needed to run {testname}")

    def addSuccess(self, test):
        super().addSuccess(test)
        self._print_time(test)

    def addError(self, test, err, traceback):
        super().addError(test, err, traceback)
        self._print_time(test)

    def addFailure(self, test, err, traceback):
        super().addFailure(test, err, traceback)
        self._print_time(test)

    def addSkip(self, test, reason):
        super().addSkip(test, reason)
        self._print_time(test)

    def startTest(self, test):
        super().startTest(test)
        self._start_time_map[
            self.format.test_name(test)
        ] = datetime.datetime.now()

    def addExpectedFailure(self, test, err):
        super().addExpectedFailure(test, err)
        self._print_time(test)

    def addUnexpectedSuccess(self, test):
        super().addUnexpectedSuccess(test)
        self._print_time(test)


class ImprovedVerboseWriter(StandardVerboseWriter):
    def __init__(
        self,
        stream,
        format_,
        descriptions,
        traceback_highlight=False,
        fast_info=False,
    ):
        super().__init__(
            stream, format_, descriptions, traceback_highlight, fast_info
        )
        self.last_test = None

    def __is_new_module(self, test):
        return (
            not self.last_test
            or test.__class__.__module__ != self.last_test.__class__.__module__
        )

    def __is_new_class(self, test):
        return (
            self.__is_new_module(test)
            or test.__class__.__name__ != self.last_test.__class__.__name__
        )

    def __format_module(self, test):
        if not self.__is_new_module(test):
            return self.format.lightgrey(test.__class__.__module__)
        return self.format.module(test)

    def __format_class(self, test):
        if not self.__is_new_class(test):
            return self.format.lightgrey(test.__class__.__name__)
        return test.__class__.__name__

    def startTest(self, test):
        self.stream.write(
            self.__format_module(test)
            + self.format.lightgrey(".")
            + self.__format_class(test)
            + self.format.lightgrey(".")
            + self.format.test_method_name(test)
            + self.format.lightgrey(" : ")
        )
        self.stream.flush()
        self.last_test = test



# pylint: disable=too-many-lines
import json
from unittest import (
    TestCase,
    mock,
)

from pcs import settings
from pcs.common import (
    communication,
    reports,
)
from pcs.common.interface import dto
from pcs.common.tools import timeout_to_seconds
from pcs.lib.commands import stonith

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import get_test_resource as rc

from .cluster.common import (
    corosync_conf_fixture,
    get_two_node,
    node_fixture,
)

STONITH_ID_SCSI = "scsi-fence-device"
STONITH_ID_MPATH = "mpath-fence-device"
STONITH_TYPE_SCSI = "fence_scsi"
STONITH_TYPE_MPATH = "fence_mpath"
STONITH_TYPE_UNSUPPORTED = "fence_unsupported"
SCSI_NODE = "node1"
_DIGEST = "0" * 31
DEFAULT_DIGEST = _DIGEST + "0"
ALL_DIGEST = _DIGEST + "1"
NONPRIVATE_DIGEST = _DIGEST + "2"
NONRELOADABLE_DIGEST = _DIGEST + "3"
DEV_1 = "/dev/sda"
DEV_2 = "/dev/sdb"
DEV_3 = "/dev/sdc"
DEV_4 = "/dev/sdd"
DEVICES_1 = ("/dev/sda",)
DEVICES_2 = ("/dev/sda", "/dev/sdb")
DEVICES_3 = ("/dev/sda", "/dev/sdb", "/dev/sdc")

DEFAULT_MONITOR = ("monitor", "60s", None, None)
DEFAULT_OPS = (DEFAULT_MONITOR,)
DEFAULT_LRM_START_OPS = (("0", DEFAULT_DIGEST, None, None),)
DEFAULT_LRM_MONITOR_OPS = (("60000", DEFAULT_DIGEST, None, None),)
DEFAULT_LRM_START_OPS_UPDATED = (("0", ALL_DIGEST, None, None),)
DEFAULT_LRM_MONITOR_OPS_UPDATED = (("60000", ALL_DIGEST, None, None),)
DEFAULT_PCMK_HOST_MAP = "node1:1;node2:2;node3:3"
DEFAULT_NODE_KEY_MAP = {"node1": "1", "node2": "2", "node3": "3"}


def _fixture_ops(resource_id, ops):
    return "\n".join(
        [
            (
                '<op id="{resource_id}-{name}-interval-{_interval}"'
                ' interval="{interval}" {timeout} name="{name}"/>'
            ).format(
                resource_id=resource_id,
                name=name,
                _interval=_interval if _interval else interval,
                interval=interval,
                timeout=f'timeout="{timeout}"' if timeout else "",
            )
            for name, interval, timeout, _interval in ops
        ]
    )


def _fixture_devices_nvpair(resource_id, devices):
    if devices is None:
        return ""
    return (
        '<nvpair id="{resource_id}-instance_attributes-devices" name="devices"'
        ' value="{devices}"/>'
    ).format(resource_id=resource_id, devices=",".join(sorted(devices)))


def fixture_scsi(
    stonith_id=STONITH_ID_SCSI,
    stonith_type=STONITH_TYPE_SCSI,
    devices=DEVICES_1,
    resource_ops=DEFAULT_OPS,
    host_map=DEFAULT_PCMK_HOST_MAP,
):
    return """
        <resources>
            <primitive class="stonith" id="{stonith_id}" type="{stonith_type}">
                <instance_attributes id="{stonith_id}-instance_attributes">
                    {devices}
                    <nvpair id="{stonith_id}-instance_attributes-pcmk_host_check" name="pcmk_host_check" value="static-list"/>
                    <nvpair id="{stonith_id}-instance_attributes-pcmk_host_list" name="pcmk_host_list" value="node1 node2 node3"/>
                    {host_map}
                    <nvpair id="{stonith_id}-instance_attributes-pcmk_reboot_action" name="pcmk_reboot_action" value="off"/>
                </instance_attributes>
                <meta_attributes id="{stonith_id}-meta_attributes">
                    <nvpair id="{stonith_id}-meta_attributes-provides" name="provides" value="unfencing"/>
                </meta_attributes>
                <operations>
                    {operations}
                </operations>
            </primitive>
            <primitive class="ocf" id="dummy" provider="pacemaker" type="Dummy"/>
        </resources>
    """.format(
        stonith_id=stonith_id,
        stonith_type=stonith_type,
        devices=_fixture_devices_nvpair(stonith_id, devices),
        operations=_fixture_ops(stonith_id, resource_ops),
        host_map=(
            (
                f'<nvpair id="{stonith_id}-instance_attributes-pcmk_host_map" '
                f'name="pcmk_host_map" value="{host_map}"/>'
            )
            if host_map is not None
            else ""
        ),
    )


def _fixture_lrm_rsc_ops(op_type, resource_id, lrm_ops):
    return [
        (
            '<lrm_rsc_op id="{resource_id}_{op_type_id}_{ms}" operation="{op_type}" '
            'interval="{ms}" {_all} {secure} {restart}/>'
        ).format(
            op_type_id="last" if op_type == "start" else op_type,
            op_type=op_type,
            resource_id=resource_id,
            ms=ms,
            _all=f'op-digest="{_all}"' if _all else "",
            secure=f'op-secure-digest="{secure}"' if secure else "",
            restart=f'op-restart-digest="{restart}"' if restart else "",
        )
        for ms, _all, secure, restart in lrm_ops
    ]


def _fixture_lrm_rsc_monitor_ops(resource_id, lrm_monitor_ops):
    return _fixture_lrm_rsc_ops("monitor", resource_id, lrm_monitor_ops)


def _fixture_lrm_rsc_start_ops(resource_id, lrm_start_ops):
    return _fixture_lrm_rsc_ops("start", resource_id, lrm_start_ops)


def _fixture_status_lrm_ops_base(
    resource_id,
    resource_type,
    lrm_ops,
):
    return f"""
        <status>
            <node_state id="1" uname="node1">
                <lrm id="1">
                    <lrm_resources>
                        <lrm_resource id="{resource_id}" type="{resource_type}" class="stonith">
                            {lrm_ops}
                        </lrm_resource>
                    </lrm_resources>
                </lrm>
            </node_state>
        </status>
    """


def _fixture_status_lrm_ops(
    resource_id,
    resource_type,
    lrm_start_ops=DEFAULT_LRM_START_OPS,
    lrm_monitor_ops=DEFAULT_LRM_MONITOR_OPS,
):
    return _fixture_status_lrm_ops_base(
        resource_id,
        resource_type,
        "\n".join(
            _fixture_lrm_rsc_start_ops(resource_id, lrm_start_ops)
            + _fixture_lrm_rsc_monitor_ops(resource_id, lrm_monitor_ops)
        ),
    )


def fixture_digests_xml(resource_id, node_name, devices=""):
    return f"""
        <pacemaker-result api-version="2.9" request="crm_resource --digests --resource {resource_id} --node {node_name} --output-as xml devices={devices}">
            <digests resource="{resource_id}" node="{node_name}" task="stop" interval="0ms">
                <digest type="all" hash="{ALL_DIGEST}">
                    <parameters devices="{devices}" pcmk_host_check="static-list" pcmk_host_list="node1 node2 node3" pcmk_reboot_action="off"/>
                </digest>
                <digest type="nonprivate" hash="{NONPRIVATE_DIGEST}">
                    <parameters devices="{devices}"/>
                </digest>
            </digests>
            <status code="0" message="OK"/>
        </pacemaker-result>
    """


def fixture_crm_mon_res_running(
    resource_id, resource_agent, nodes_running_on=1
):
    role = "Started" if nodes_running_on else "Stopped"
    nodes = "\n".join(
        f'<node name="node{i}" id="{i}" cached="true"/>'
        for i in range(1, nodes_running_on + 1)
    )
    return f"""
        <resources>
            <resource id="{resource_id}" resource_agent="{resource_agent}"
                role="{role}" nodes_running_on="{nodes_running_on}">
                {nodes}
            </resource>
        </resources>
    """


FIXTURE_CRM_MON_NODES = """
    <nodes>
        <node name="node1" id="1" is_dc="true" resources_running="1"/>
        <node name="node2" id="2"/>
        <node name="node3" id="3"/>
    </nodes>
"""


FIXTURE_CRM_MON_RES_STOPPED = f"""
    <resource id="{STONITH_ID_SCSI}" resource_agent="stonith:fence_scsi" role="Stopped" nodes_running_on="0"/>
"""


class CommandSetMixin:
    def command(self, **kwargs):
        return lambda: stonith.update_scsi_devices(
            self.env_assist.get_env(),
            self.stonith_id,
            kwargs.get("devices_updated", DEVICES_2),
            force_flags=kwargs.get("force_flags", ()),
        )


class CommandAddRemoveMixin:
    def command(self, **kwargs):
        return lambda: stonith.update_scsi_devices_add_remove(
            self.env_assist.get_env(),
            self.stonith_id,
            kwargs.get("devices_add", [DEV_2]),
            kwargs.get("devices_remove", ()),
            force_flags=kwargs.get("force_flags", ()),
        )


class MpathMixin:
    command_url = "api/v1/scsi-unfence-node-mpath/v1"
    stonith_id = STONITH_ID_MPATH
    stonith_type = STONITH_TYPE_MPATH

    def unfence_config(
        self,
        original_devices=(),
        updated_devices=(),
        node_labels=None,
        communication_list=None,
    ):
        self.config.http.scsi.unfence_node_mpath(
            DEFAULT_NODE_KEY_MAP,
            original_devices=original_devices,
            updated_devices=updated_devices,
            node_labels=node_labels,
            communication_list=communication_list,
        )

    @staticmethod
    def get_raw_data(node):
        return dict(
            key=DEFAULT_NODE_KEY_MAP[node],
            original_devices=DEVICES_1,
            updated_devices=DEVICES_2,
        )


class ScsiMixin:
    command_url = "api/v1/scsi-unfence-node/v2"
    stonith_id = STONITH_ID_SCSI
    stonith_type = STONITH_TYPE_SCSI

    def unfence_config(
        self,
        original_devices=(),
        updated_devices=(),
        node_labels=None,
        communication_list=None,
    ):
        self.config.http.scsi.unfence_node(
            original_devices=original_devices,
            updated_devices=updated_devices,
            node_labels=node_labels,
            communication_list=communication_list,
        )

    @staticmethod
    def get_raw_data(node):
        return dict(
            node=node, original_devices=DEVICES_1, updated_devices=DEVICES_2
        )


class UpdateScsiDevicesMixin:
    def setUp(self):
        # pylint: disable=invalid-name
        self.env_assist, self.config = get_env_tools(self)

        self.existing_nodes = ["node1", "node2", "node3"]
        self.existing_corosync_nodes = [
            node_fixture(node, node_id)
            for node_id, node in enumerate(self.existing_nodes, 1)
        ]
        self.config.env.set_known_nodes(self.existing_nodes)

    def config_cib(
        self,
        devices_before=DEVICES_1,
        devices_updated=DEVICES_2,
        resource_ops=DEFAULT_OPS,
        lrm_monitor_ops=DEFAULT_LRM_MONITOR_OPS,
        lrm_start_ops=DEFAULT_LRM_START_OPS,
        host_map=DEFAULT_PCMK_HOST_MAP,
        nodes_running_on=1,
        start_digests=True,
        monitor_digests=True,
    ):
        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-locals
        devices_value = ",".join(sorted(devices_updated))
        self.config.runner.cib.load(
            resources=fixture_scsi(
                stonith_id=self.stonith_id,
                stonith_type=self.stonith_type,
                devices=devices_before,
                resource_ops=resource_ops,
                host_map=host_map,
            ),
            status=_fixture_status_lrm_ops(
                self.stonith_id,
                self.stonith_type,
                lrm_start_ops=lrm_start_ops,
                lrm_monitor_ops=lrm_monitor_ops,
            ),
        )
        self.config.runner.pcmk.is_resource_digests_supported()
        self.config.runner.pcmk.load_state(
            resources=fixture_crm_mon_res_running(
                self.stonith_id,
                f"stonith:{self.stonith_type}",
                nodes_running_on=nodes_running_on,
            ),
            nodes=FIXTURE_CRM_MON_NODES,
        )
        devices_opt = "devices={}".format(devices_value)
        if start_digests:
            self.config.runner.pcmk.resource_digests(
                self.stonith_id,
                SCSI_NODE,
                name="start.op.digests",
                stdout=fixture_digests_xml(
                    self.stonith_id, SCSI_NODE, devices=devices_value
                ),
                args=[devices_opt],
            )
        if monitor_digests:
            for num, op in enumerate(resource_ops, 1):
                name, interval, timeout, _ = op
                if name != "monitor":
                    continue
                args = [devices_opt]
                args.append(
                    "CRM_meta_interval={}".format(
                        1000 * timeout_to_seconds(interval)
                    )
                )
                if timeout:
                    args.append(
                        "CRM_meta_timeout={}".format(
                            1000 * timeout_to_seconds(timeout)
                        )
                    )
                self.config.runner.pcmk.resource_digests(
                    self.stonith_id,
                    SCSI_NODE,
                    name=f"{name}-{num}.op.digests",
                    stdout=fixture_digests_xml(
                        self.stonith_id,
                        SCSI_NODE,
                        devices=devices_value,
                    ),
                    args=args,
                )

    def assert_command_success(
        self,
        devices_before=DEVICES_1,
        devices_updated=DEVICES_2,
        devices_add=(),
        devices_remove=(),
        unfence=None,
        resource_ops=DEFAULT_OPS,
        lrm_monitor_ops=DEFAULT_LRM_MONITOR_OPS,
        lrm_start_ops=DEFAULT_LRM_START_OPS,
        lrm_monitor_ops_updated=DEFAULT_LRM_MONITOR_OPS_UPDATED,
        lrm_start_ops_updated=DEFAULT_LRM_START_OPS_UPDATED,
    ):
        # pylint: disable=too-many-arguments
        self.config_cib(
            devices_before=devices_before,
            devices_updated=devices_updated,
            resource_ops=resource_ops,
            lrm_monitor_ops=lrm_monitor_ops,
            lrm_start_ops=lrm_start_ops,
        )
        if unfence:
            self.config.corosync_conf.load_content(
                corosync_conf_fixture(
                    self.existing_corosync_nodes,
                    get_two_node(len(self.existing_corosync_nodes)),
                )
            )
            self.config.http.corosync.get_corosync_online_targets(
                node_labels=self.existing_nodes
            )
            self.unfence_config(
                original_devices=devices_before,
                updated_devices=devices_updated,
                node_labels=self.existing_nodes,
            )
        self.config.env.push_cib(
            resources=fixture_scsi(
                stonith_id=self.stonith_id,
                stonith_type=self.stonith_type,
                devices=devices_updated,
                resource_ops=resource_ops,
            ),
            status=_fixture_status_lrm_ops(
                self.stonith_id,
                self.stonith_type,
                lrm_start_ops=lrm_start_ops_updated,
                lrm_monitor_ops=lrm_monitor_ops_updated,
            ),
        )
        self.command(
            devices_updated=devices_updated,
            devices_add=devices_add,
            devices_remove=devices_remove,
        )()
        self.env_assist.assert_reports([])


class UpdateScsiDevicesFailuresMixin(UpdateScsiDevicesMixin):
    def test_pcmk_doesnt_support_digests(self):
        self.config.runner.cib.load(
            resources=fixture_scsi(
                stonith_id=self.stonith_id, stonith_type=self.stonith_type
            )
        )
        self.config.runner.pcmk.is_resource_digests_supported(
            is_supported=False
        )
        self.env_assist.assert_raise_library_error(
            self.command(),
            [
                fixture.error(
                    reports.codes.STONITH_RESTARTLESS_UPDATE_OF_SCSI_DEVICES_NOT_SUPPORTED,
                )
            ],
            expected_in_processor=False,
        )

    def test_nonexistent_id(self):
        """
        lower level tested in
        pcs_test.tier0.lib.cib.test_stonith.ValidateStonithRestartlessUpdate
        """
        self.config.runner.cib.load(
            resources=fixture_scsi(
                stonith_id="not-searched-id",
                stonith_type=self.stonith_type,
            )
        )
        self.config.runner.pcmk.is_resource_digests_supported()
        self.env_assist.assert_raise_library_error(self.command())
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ID_NOT_FOUND,
                    id=self.stonith_id,
                    expected_types=["primitive"],
                    context_type="cib",
                    context_id="",
                )
            ]
        )

    def test_not_supported_resource_type(self):
        """
        lower level tested in
        pcs_test.tier0.lib.cib.test_stonith.ValidateStonithRestartlessUpdate
        """
        self.config.runner.cib.load(
            resources=fixture_scsi(
                stonith_id=self.stonith_id,
                stonith_type=STONITH_TYPE_UNSUPPORTED,
            )
        )
        self.config.runner.pcmk.is_resource_digests_supported()
        self.env_assist.assert_raise_library_error(
            self.command(devices_add=[DEV_2], devices_remove=[DEV_1])
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.STONITH_RESTARTLESS_UPDATE_UNSUPPORTED_AGENT,
                    resource_id=self.stonith_id,
                    resource_type=STONITH_TYPE_UNSUPPORTED,
                    supported_stonith_types=["fence_scsi", "fence_mpath"],
                )
            ]
        )

    def test_stonith_resource_is_not_running(self):
        self.config_cib(
            nodes_running_on=0, start_digests=False, monitor_digests=False
        )
        self.env_assist.assert_raise_library_error(
            self.command(),
            [
                fixture.error(
                    reports.codes.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM,
                    reason=f"resource '{self.stonith_id}' is not running on any node",
                    reason_type=reports.const.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM_REASON_NOT_RUNNING,
                )
            ],
            expected_in_processor=False,
        )

    def test_stonith_resource_is_running_on_more_than_one_node(self):
        self.config_cib(
            nodes_running_on=2, start_digests=False, monitor_digests=False
        )
        self.env_assist.assert_raise_library_error(
            self.command(),
            [
                fixture.error(
                    reports.codes.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM,
                    reason=(
                        f"resource '{self.stonith_id}' is running on more than "
                        "1 node"
                    ),
                    reason_type=reports.const.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM_REASON_OTHER,
                )
            ],
            expected_in_processor=False,
        )

    def test_no_lrm_start_op(self):
        self.config_cib(
            lrm_start_ops=(), start_digests=False, monitor_digests=False
        )
        self.env_assist.assert_raise_library_error(
            self.command(),
            [
                fixture.error(
                    reports.codes.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM,
                    reason=(
                        "lrm_rsc_op element for start operation was not found"
                    ),
                    reason_type=reports.const.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM_REASON_OTHER,
                )
            ],
            expected_in_processor=False,
        )

    def test_lrm_op_missing_digest_attributes(self):
        self.config_cib(
            resource_ops=(),
            lrm_start_ops=(("0", None, None, None),),
            lrm_monitor_ops=(),
        )
        self.env_assist.assert_raise_library_error(
            self.command(),
            [
                fixture.error(
                    reports.codes.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM,
                    reason="no digests attributes in lrm_rsc_op element",
                    reason_type=reports.const.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM_REASON_OTHER,
                )
            ],
            expected_in_processor=False,
        )

    def test_crm_resource_digests_missing(self):
        self.config_cib(
            lrm_start_ops=(("0", None, None, "somedigest"),),
            monitor_digests=False,
        )
        self.env_assist.assert_raise_library_error(
            self.command(),
            [
                fixture.error(
                    reports.codes.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM,
                    reason=(
                        "necessary digest for 'op-restart-digest' attribute is "
                        "missing"
                    ),
                    reason_type=reports.const.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM_REASON_OTHER,
                )
            ],
            expected_in_processor=False,
        )

    def test_monitor_ops_and_lrm_monitor_ops_do_not_match(self):
        self.config_cib(
            resource_ops=(
                ("monitor", "30s", "10s", None),
                ("monitor", "30s", "20s", "31"),
                ("monitor", "60s", None, None),
            ),
            monitor_digests=False,
        )
        self.env_assist.assert_raise_library_error(
            self.command(),
            [
                fixture.error(
                    reports.codes.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM,
                    reason=(
                        "number of lrm_rsc_op and op elements for monitor "
                        "operation differs"
                    ),
                    reason_type=reports.const.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM_REASON_OTHER,
                )
            ],
            expected_in_processor=False,
        )

    def test_lrm_monitor_ops_not_found(self):
        self.config_cib(
            resource_ops=(("monitor", "30s", None, None),),
            monitor_digests=False,
        )
        self.env_assist.assert_raise_library_error(
            self.command(),
            [
                fixture.error(
                    reports.codes.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM,
                    reason=(
                        "monitor lrm_rsc_op element for resource "
                        f"'{self.stonith_id}', node '{SCSI_NODE}' and interval "
                        "'30000' not found"
                    ),
                    reason_type=reports.const.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM_REASON_OTHER,
                )
            ],
            expected_in_processor=False,
        )

    def test_node_missing_name_and_missing_auth_token(self):
        self.config_cib()
        self.config.corosync_conf.load_content(
            corosync_conf_fixture(
                self.existing_corosync_nodes
                + [[("ring0_addr", "custom_node"), ("nodeid", "5")]],
            )
        )
        self.config.env.set_known_nodes(self.existing_nodes[:-1])
        self.env_assist.assert_raise_library_error(self.command())
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                    fatal=True,
                ),
                fixture.error(
                    reports.codes.HOST_NOT_FOUND,
                    host_list=[self.existing_nodes[-1]],
                ),
            ]
        )

    def test_unfence_failure_unable_to_connect(self):
        self.config_cib()
        self.config.corosync_conf.load_content(
            corosync_conf_fixture(self.existing_corosync_nodes)
        )
        self.config.http.corosync.get_corosync_online_targets(
            node_labels=self.existing_nodes
        )
        communication_list = [
            dict(label=node, raw_data=json.dumps(self.get_raw_data(node)))
            for node in self.existing_nodes
        ]
        communication_list[0]["was_connected"] = False
        communication_list[0]["error_msg"] = "errA"
        communication_list[1]["output"] = json.dumps(
            dto.to_dict(
                communication.dto.InternalCommunicationResultDto(
                    status=communication.const.COM_STATUS_ERROR,
                    status_msg="error",
                    report_list=[
                        reports.ReportItem.error(
                            reports.messages.StonithUnfencingFailed("errB")
                        ).to_dto()
                    ],
                    data=None,
                )
            )
        )
        self.unfence_config(communication_list=communication_list)
        self.env_assist.assert_raise_library_error(self.command())
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.existing_nodes[0],
                    command=self.command_url,
                    reason="errA",
                ),
                fixture.error(
                    reports.codes.STONITH_UNFENCING_FAILED,
                    reason="errB",
                    context=reports.dto.ReportItemContextDto(
                        node=self.existing_nodes[1],
                    ),
                ),
            ]
        )

    def test_corosync_targets_unable_to_connect(self):
        self.config_cib()
        self.config.corosync_conf.load_content(
            corosync_conf_fixture(self.existing_corosync_nodes)
        )
        self.config.http.corosync.get_corosync_online_targets(
            communication_list=[
                dict(
                    label=self.existing_nodes[0],
                    output='{"corosync":true}',
                ),
            ]
            + [
                dict(
                    label=node,
                    was_connected=False,
                    errno=7,
                    error_msg="an error",
                )
                for node in self.existing_nodes[1:]
            ]
        )
        self.env_assist.assert_raise_library_error(self.command())
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    force_code=reports.codes.SKIP_OFFLINE_NODES,
                    node=node,
                    command="remote/status",
                    reason="an error",
                )
                for node in self.existing_nodes[1:]
            ]
        )

    def test_corosync_targets_skip_offline_unfence_node_running_corosync(
        self,
    ):
        self.config_cib()
        self.config.corosync_conf.load_content(
            corosync_conf_fixture(self.existing_corosync_nodes)
        )
        self.config.http.corosync.get_corosync_online_targets(
            communication_list=[
                dict(
                    label=self.existing_nodes[0],
                    output='{"corosync":true}',
                ),
                dict(
                    label=self.existing_nodes[1],
                    output='{"corosync":false}',
                ),
                dict(
                    label=self.existing_nodes[2],
                    was_connected=False,
                    errno=7,
                    error_msg="an error",
                ),
            ]
        )
        communication_list = [
            dict(
                label=self.existing_nodes[0],
                raw_data=json.dumps(self.get_raw_data(self.existing_nodes[0])),
            ),
        ]
        self.unfence_config(communication_list=communication_list)
        self.config.env.push_cib(
            resources=fixture_scsi(
                stonith_id=self.stonith_id,
                stonith_type=self.stonith_type,
                devices=DEVICES_2,
            ),
            status=_fixture_status_lrm_ops(
                self.stonith_id,
                self.stonith_type,
                lrm_start_ops=DEFAULT_LRM_START_OPS_UPDATED,
                lrm_monitor_ops=DEFAULT_LRM_MONITOR_OPS_UPDATED,
            ),
        )
        self.command(force_flags=[reports.codes.SKIP_OFFLINE_NODES])()
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.existing_nodes[2],
                    command="remote/status",
                    reason="an error",
                ),
            ]
        )

    def test_corosync_targets_unable_to_perform_unfencing_operation(
        self,
    ):
        self.config_cib()
        self.config.corosync_conf.load_content(
            corosync_conf_fixture(self.existing_corosync_nodes)
        )
        self.config.http.corosync.get_corosync_online_targets(
            communication_list=[
                dict(
                    label=self.existing_nodes[0],
                    was_connected=False,
                    errno=7,
                    error_msg="an error",
                ),
                dict(
                    label=self.existing_nodes[1],
                    was_connected=False,
                    errno=7,
                    error_msg="an error",
                ),
                dict(
                    label=self.existing_nodes[2],
                    output='{"corosync":false}',
                ),
            ]
        )
        self.unfence_config(communication_list=[])
        self.env_assist.assert_raise_library_error(
            self.command(force_flags=[reports.codes.SKIP_OFFLINE_NODES])
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=node,
                    command="remote/status",
                    reason="an error",
                )
                for node in self.existing_nodes[0:2]
            ]
            + [
                fixture.error(
                    reports.codes.UNABLE_TO_PERFORM_OPERATION_ON_ANY_NODE,
                ),
            ]
        )

    def test_unfence_failure_unknown_command(self):
        self.config_cib()
        self.config.corosync_conf.load_content(
            corosync_conf_fixture(self.existing_corosync_nodes)
        )
        self.config.http.corosync.get_corosync_online_targets(
            node_labels=self.existing_nodes
        )
        communication_list = [
            dict(
                label=node,
                raw_data=json.dumps(self.get_raw_data(node)),
            )
            for node in self.existing_nodes
        ]
        communication_list[2]["response_code"] = 404
        communication_list[2]["output"] = json.dumps(
            dto.to_dict(
                communication.dto.InternalCommunicationResultDto(
                    status=communication.const.COM_STATUS_UNKNOWN_CMD,
                    status_msg=("Unknown command '/api/v1/unknown/v2'"),
                    report_list=[],
                    data=None,
                )
            )
        )
        self.unfence_config(communication_list=communication_list)
        self.env_assist.assert_raise_library_error(self.command())
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.PCSD_VERSION_TOO_OLD,
                    node=self.existing_nodes[2],
                ),
            ]
        )

    def test_unfence_failure_agent_script_failed(self):
        self.config_cib()
        self.config.corosync_conf.load_content(
            corosync_conf_fixture(self.existing_corosync_nodes)
        )
        self.config.http.corosync.get_corosync_online_targets(
            node_labels=self.existing_nodes
        )
        communication_list = [
            dict(
                label=node,
                raw_data=json.dumps(self.get_raw_data(node)),
            )
            for node in self.existing_nodes
        ]
        communication_list[1]["output"] = json.dumps(
            dto.to_dict(
                communication.dto.InternalCommunicationResultDto(
                    status=communication.const.COM_STATUS_ERROR,
                    status_msg="error",
                    report_list=[
                        reports.ReportItem.error(
                            reports.messages.StonithUnfencingFailed("errB")
                        ).to_dto()
                    ],
                    data=None,
                )
            )
        )
        self.unfence_config(communication_list=communication_list)
        self.env_assist.assert_raise_library_error(self.command())
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.STONITH_UNFENCING_FAILED,
                    reason="errB",
                    context=reports.dto.ReportItemContextDto(
                        node=self.existing_nodes[1],
                    ),
                ),
            ]
        )


class UpdateScsiDevicesSetBase(UpdateScsiDevicesMixin, CommandSetMixin):
    def test_update_1_to_1_devices(self):
        self.assert_command_success(
            devices_before=DEVICES_1,
            devices_updated=DEVICES_1,
        )

    def test_update_2_to_2_devices(self):
        self.assert_command_success(
            devices_before=DEVICES_2,
            devices_updated=DEVICES_2,
        )

    def test_update_1_to_2_devices(self):
        self.assert_command_success(unfence=[DEV_2])

    def test_update_1_to_3_devices(self):
        self.assert_command_success(
            devices_before=DEVICES_1,
            devices_updated=DEVICES_3,
            unfence=[DEV_2, DEV_3],
        )

    def test_update_3_to_1_devices(self):
        self.assert_command_success(
            devices_before=DEVICES_3,
            devices_updated=DEVICES_1,
        )

    def test_update_3_to_2_devices(self):
        self.assert_command_success(
            devices_before=DEVICES_3,
            devices_updated=DEVICES_2,
        )

    def test_update_add_2_to_2_remove_1(self):
        self.assert_command_success(
            devices_before=[DEV_1, DEV_2],
            devices_updated=[DEV_2, DEV_3, DEV_4],
            unfence=[DEV_3, DEV_4],
        )

    def test_default_monitor(self):
        self.assert_command_success(unfence=[DEV_2])

    def test_no_monitor_ops(self):
        self.assert_command_success(
            unfence=[DEV_2],
            resource_ops=(),
            lrm_monitor_ops=(),
            lrm_monitor_ops_updated=(),
        )

    def test_1_monitor_with_timeout(self):
        self.assert_command_success(
            unfence=[DEV_2],
            resource_ops=(("monitor", "30s", "10s", None),),
            lrm_monitor_ops=(("30000", DEFAULT_DIGEST, None, None),),
            lrm_monitor_ops_updated=(("30000", ALL_DIGEST, None, None),),
        )

    def test_2_monitor_ops_with_timeouts(self):
        self.assert_command_success(
            unfence=[DEV_2],
            resource_ops=(
                ("monitor", "30s", "10s", None),
                ("monitor", "40s", "20s", None),
            ),
            lrm_monitor_ops=(
                ("30000", DEFAULT_DIGEST, None, None),
                ("40000", DEFAULT_DIGEST, None, None),
            ),
            lrm_monitor_ops_updated=(
                ("30000", ALL_DIGEST, None, None),
                ("40000", ALL_DIGEST, None, None),
            ),
        )

    def test_2_monitor_ops_with_one_timeout(self):
        self.assert_command_success(
            unfence=[DEV_2],
            resource_ops=(
                ("monitor", "30s", "10s", None),
                ("monitor", "60s", None, None),
            ),
            lrm_monitor_ops=(
                ("30000", DEFAULT_DIGEST, None, None),
                ("60000", DEFAULT_DIGEST, None, None),
            ),
            lrm_monitor_ops_updated=(
                ("30000", ALL_DIGEST, None, None),
                ("60000", ALL_DIGEST, None, None),
            ),
        )

    def test_various_start_ops_one_lrm_start_op(self):
        self.assert_command_success(
            unfence=[DEV_2],
            resource_ops=(
                ("monitor", "60s", None, None),
                ("start", "0s", "40s", None),
                ("start", "0s", "30s", "1"),
                ("start", "10s", "5s", None),
                ("start", "20s", None, None),
            ),
        )

    def test_1_nonrecurring_start_op_with_timeout(self):
        self.assert_command_success(
            unfence=[DEV_2],
            resource_ops=(
                ("monitor", "60s", None, None),
                ("start", "0s", "40s", None),
            ),
        )


class UpdateScsiDevicesAddRemoveBase(
    UpdateScsiDevicesMixin, CommandAddRemoveMixin
):
    def test_add_1_to_1(self):
        self.assert_command_success(
            devices_before=[DEV_1],
            devices_updated=[DEV_1, DEV_2],
            devices_add=[DEV_2],
            devices_remove=[],
            unfence=[DEV_2],
        )

    def test_add_2_to_1(self):
        self.assert_command_success(
            devices_before=[DEV_1],
            devices_updated=[DEV_1, DEV_2, DEV_3],
            devices_add=[DEV_2, DEV_3],
            devices_remove=[],
            unfence=[DEV_2, DEV_3],
        )

    def test_add_2_to_2_and_remove_1(self):
        self.assert_command_success(
            devices_before=[DEV_1, DEV_2],
            devices_updated=[DEV_2, DEV_3, DEV_4],
            devices_add=[DEV_3, DEV_4],
            devices_remove=[DEV_1],
            unfence=[DEV_3, DEV_4],
        )

    def test_remove_1_from_2(self):
        self.assert_command_success(
            devices_before=[DEV_1, DEV_2],
            devices_updated=[DEV_2],
            devices_add=[],
            devices_remove=[DEV_1],
        )

    def test_remove_2_from_3(self):
        self.assert_command_success(
            devices_before=[DEV_1, DEV_2, DEV_3],
            devices_updated=[DEV_3],
            devices_add=[],
            devices_remove=[DEV_2, DEV_1],
        )

    def test_remove_2_from_3_add_1(self):
        self.assert_command_success(
            devices_before=[DEV_1, DEV_2, DEV_3],
            devices_updated=[DEV_3, DEV_4],
            devices_add=[DEV_4],
            devices_remove=[DEV_2, DEV_1],
            unfence=[DEV_4],
        )

    def test_add_1_remove_1(self):
        self.assert_command_success(
            devices_before=[DEV_1, DEV_2],
            devices_updated=[DEV_2, DEV_3],
            devices_add=[DEV_3],
            devices_remove=[DEV_1],
            unfence=[DEV_3],
        )

    def test_add_2_remove_2(self):
        self.assert_command_success(
            devices_before=[DEV_1, DEV_2],
            devices_updated=[DEV_3, DEV_4],
            devices_add=[DEV_3, DEV_4],
            devices_remove=[DEV_1, DEV_2],
            unfence=[DEV_3, DEV_4],
        )


class UpdateScsiDevicesSetFailuresBaseMixin(
    UpdateScsiDevicesFailuresMixin, CommandSetMixin
):
    def test_devices_cannot_be_empty(self):
        self.config.runner.cib.load(
            resources=fixture_scsi(
                stonith_id=self.stonith_id, stonith_type=self.stonith_type
            )
        )
        self.config.runner.pcmk.is_resource_digests_supported()
        self.env_assist.assert_raise_library_error(
            self.command(devices_updated=())
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="devices",
                    option_value="",
                    allowed_values=None,
                    cannot_be_empty=True,
                    forbidden_characters=None,
                )
            ]
        )


class UpdateScsiDevicesAddRemoveFailuresBaseMixin(
    UpdateScsiDevicesFailuresMixin, CommandAddRemoveMixin
):
    def test_add_remove_are_empty(self):
        """
        lower level tested in
        pcs_test/tier0/lib/test_validate.ValidateAddRemoveItems
        """
        self.config.runner.cib.load(
            resources=fixture_scsi(
                stonith_id=self.stonith_id, stonith_type=self.stonith_type
            )
        )
        self.config.runner.pcmk.is_resource_digests_supported()
        self.env_assist.assert_raise_library_error(
            self.command(devices_add=(), devices_remove=())
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_ITEMS_NOT_SPECIFIED,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_STONITH_RESOURCE,
                    item_type="device",
                    container_id=self.stonith_id,
                )
            ]
        )


class MpathFailuresMixin:
    def assert_failure(self, host_map, missing_nodes=None):
        self.config_cib(host_map=host_map)
        self.config.corosync_conf.load_content(
            corosync_conf_fixture(
                self.existing_corosync_nodes,
                get_two_node(len(self.existing_corosync_nodes)),
            )
        )
        self.config.http.corosync.get_corosync_online_targets(
            node_labels=self.existing_nodes
        )
        if missing_nodes is None:
            missing_nodes = []
        if not host_map:
            host_map = None
        self.env_assist.assert_raise_library_error(
            self.command(),
            [
                fixture.error(
                    reports.codes.STONITH_RESTARTLESS_UPDATE_MISSING_MPATH_KEYS,
                    pcmk_host_map_value=host_map,
                    missing_nodes=sorted(missing_nodes),
                )
            ],
            expected_in_processor=False,
        )

    def test_host_map_missing(self):
        self.assert_failure(None, self.existing_nodes)

    def test_host_map_empty(self):
        self.assert_failure("", self.existing_nodes)

    def test_missing_keys_for_nodes(self):
        self.assert_failure("node1:1;node2=", ["node2", "node3"])


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_api_rng/api-result.rng"),
)
class TestUpdateScsiDevicesSetMpath(
    UpdateScsiDevicesSetBase, MpathMixin, TestCase
):
    pass


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_api_rng/api-result.rng"),
)
class TestUpdateScsiDevicesSetScsi(
    UpdateScsiDevicesSetBase, ScsiMixin, TestCase
):
    pass


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_api_rng/api-result.rng"),
)
class TestUpdateScsiDevicesAddRemoveMpath(
    UpdateScsiDevicesAddRemoveBase, MpathMixin, TestCase
):
    pass


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_api_rng/api-result.rng"),
)
class TestUpdateScsiDevicesAddRemoveScsi(
    UpdateScsiDevicesAddRemoveBase, ScsiMixin, TestCase
):
    pass


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_api_rng/api-result.rng"),
)
class TestUpdateScsiDevicesFailuresMpath(
    UpdateScsiDevicesSetFailuresBaseMixin,
    MpathFailuresMixin,
    MpathMixin,
    TestCase,
):
    pass


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_api_rng/api-result.rng"),
)
class TestUpdateScsiDevicesFailuresScsi(
    UpdateScsiDevicesSetFailuresBaseMixin, ScsiMixin, TestCase
):
    pass


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_api_rng/api-result.rng"),
)
class TestUpdateScsiDevicesAddRemoveFailuresMpath(
    UpdateScsiDevicesAddRemoveFailuresBaseMixin,
    MpathMixin,
    MpathFailuresMixin,
    TestCase,
):
    pass


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_api_rng/api-result.rng"),
)
class TestUpdateScsiDevicesAddRemoveFailuresScsi(
    UpdateScsiDevicesAddRemoveFailuresBaseMixin, ScsiMixin, TestCase
):
    pass



from collections import namedtuple
from typing import (
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    cast,
)

from lxml.etree import _Element

from pcs.common import reports
from pcs.common.reports.item import (
    ReportItem,
    ReportItemList,
)
from pcs.common.types import StringCollection
from pcs.lib.cib import nvpair
from pcs.lib.cib.resource.bundle import TAG as TAG_BUNDLE
from pcs.lib.cib.resource.bundle import (
    get_inner_resource as get_bundle_inner_resource,
)
from pcs.lib.cib.resource.bundle import is_bundle
from pcs.lib.cib.resource.clone import ALL_TAGS as TAG_CLONE_ALL
from pcs.lib.cib.resource.clone import (
    get_inner_resource as get_clone_inner_resource,
)
from pcs.lib.cib.resource.clone import (
    get_parent_any_clone,
    is_any_clone,
    is_master,
    is_promotable_clone,
)
from pcs.lib.cib.resource.group import TAG as TAG_GROUP
from pcs.lib.cib.resource.group import (
    get_inner_resources as get_group_inner_resources,
)
from pcs.lib.cib.resource.group import is_group
from pcs.lib.cib.resource.primitive import TAG as TAG_PRIMITIVE
from pcs.lib.cib.resource.primitive import is_primitive
from pcs.lib.cib.tools import (
    ElementSearcher,
    IdProvider,
)
from pcs.lib.xml_tools import find_parent

ALL_RESOURCE_XML_TAGS = sorted(
    TAG_CLONE_ALL + [TAG_GROUP, TAG_PRIMITIVE, TAG_BUNDLE]
)


def are_meta_disabled(meta_attributes: Mapping[str, str]) -> bool:
    return meta_attributes.get("target-role", "Started").lower() == "stopped"


def _can_be_evaluated_as_positive_num(value: str) -> bool:
    string_wo_leading_zeros = str(value).lstrip("0")
    return (
        bool(string_wo_leading_zeros) and string_wo_leading_zeros[0].isdigit()
    )


def is_clone_deactivated_by_meta(meta_attributes: Mapping[str, str]) -> bool:
    return are_meta_disabled(meta_attributes) or any(
        not _can_be_evaluated_as_positive_num(meta_attributes.get(key, "1"))
        for key in ["clone-max", "clone-node-max"]
    )


def find_one_resource(
    context_element: _Element,
    resource_id: str,
    resource_tags: Optional[StringCollection] = None,
) -> Tuple[Optional[_Element], ReportItemList]:
    """
    Find a single resource or return None if not found

    context_element -- an element to be searched in
    resource_id -- id of an element to find
    resource_tags -- types of resources to look for, default all types
    """
    resource_el_list, report_list = find_resources(
        context_element,
        [resource_id],
        resource_tags=resource_tags,
    )
    resource = resource_el_list[0] if resource_el_list else None
    return resource, report_list


def find_resources(
    context_element: _Element,
    resource_ids: Iterable[str],
    resource_tags: Optional[StringCollection] = None,
) -> Tuple[List[_Element], ReportItemList]:
    """
    Find a list of resources

    context_element -- an element to be searched in
    resource_id -- id of an element to find
    resource_tags -- types of resources to look for, default all types
    """
    report_list: ReportItemList = []
    if resource_tags is None:
        resource_tags = ALL_RESOURCE_XML_TAGS
    resource_el_list = []
    for res_id in resource_ids:
        searcher = ElementSearcher(resource_tags, res_id, context_element)
        if searcher.element_found():
            resource_el_list.append(searcher.get_element())
        else:
            report_list.extend(searcher.get_errors())
    return resource_el_list, report_list


def find_primitives(resource_el: _Element) -> List[_Element]:
    """
    Get list of primitives contained in a given resource

    resource_el -- resource element
    """
    if is_bundle(resource_el):
        in_bundle = get_bundle_inner_resource(resource_el)
        return [in_bundle] if in_bundle is not None else []
    if is_any_clone(resource_el):
        resource_el = get_clone_inner_resource(resource_el)
    if is_group(resource_el):
        return get_group_inner_resources(resource_el)
    if is_primitive(resource_el):
        return [resource_el]
    return []


def get_all_inner_resources(resource_el: _Element) -> Set[_Element]:
    """
    Return all inner resources (both direct and indirect) of a resource
    Example: for a clone containing a group, this function will return both
    the group and the resources inside the group

    resource_el -- resource element to get its inner resources
    """
    all_inner: Set[_Element] = set()
    to_process = set([resource_el])
    while to_process:
        new_inner = get_inner_resources(to_process.pop())
        to_process.update(set(new_inner) - all_inner)
        all_inner.update(new_inner)
    return all_inner


def get_inner_resources(resource_el: _Element) -> List[_Element]:
    """
    Return list of inner resources (direct descendants) of a resource
    specified as resource_el.
    Example: for clone containing a group, this function will return only
    group and not resource inside the group

    resource_el -- resource element to get its inner resources
    """
    if is_bundle(resource_el):
        in_bundle = get_bundle_inner_resource(resource_el)
        return [in_bundle] if in_bundle is not None else []
    if is_any_clone(resource_el):
        return [get_clone_inner_resource(resource_el)]
    if is_group(resource_el):
        return get_group_inner_resources(resource_el)
    return []


def is_resource(element: _Element) -> bool:
    """
    Return True for any resource, False otherwise, like meta_attributes element.

    element -- element to check
    """
    return element.tag in ALL_RESOURCE_XML_TAGS


def is_wrapper_resource(resource_el: _Element) -> bool:
    """
    Return True for resource_el of types that can contain other resource(s)
    (these are: group, bundle, clone) and False otherwise.

    resource_el -- resource element to check
    """
    return (
        is_group(resource_el)
        or is_bundle(resource_el)
        or is_any_clone(resource_el)
    )


def get_parent_resource(resource_el: _Element) -> Optional[_Element]:
    """
    Return a direct ancestor of a specified resource or None if the resource
    has no ancestor.
    Example: for a resource in group which is in clone, this function will
    return group element.

    resource_el -- resource element of which parent resource should be returned
    """
    parent_el = resource_el.getparent()
    if parent_el is not None and is_wrapper_resource(parent_el):
        return parent_el
    return None


def find_resources_to_enable(resource_el: _Element) -> List[_Element]:
    """
    Get resources to enable in order to enable specified resource successfully

    resource_el -- resource element
    """
    if is_bundle(resource_el):
        to_enable = [resource_el]
        in_bundle = get_bundle_inner_resource(resource_el)
        if in_bundle is not None:
            to_enable.append(in_bundle)
        return to_enable

    if is_any_clone(resource_el):
        return [resource_el, get_clone_inner_resource(resource_el)]

    to_enable = [resource_el]
    parent = resource_el.getparent()
    if parent is not None and (is_any_clone(parent) or is_bundle(parent)):
        to_enable.append(parent)
    return to_enable


def enable(resource_el: _Element, id_provider: IdProvider) -> None:
    """
    Enable specified resource

    resource_el -- resource element
    """
    nvpair.arrange_first_meta_attributes(
        resource_el,
        {"target-role": ""},
        id_provider,
    )


def disable(resource_el: _Element, id_provider: IdProvider) -> None:
    """
    Disable specified resource

    resource_el -- resource element
    """
    nvpair.arrange_first_meta_attributes(
        resource_el,
        {"target-role": "Stopped"},
        id_provider,
    )


def find_resources_to_manage(resource_el: _Element) -> List[_Element]:
    """
    Get resources to set to managed for the specified resource to become managed

    resource_el -- resource element
    """
    # If the resource_el is a primitive in a group, we set both the group and
    # the primitive to managed mode. Otherwise the resource_el, all its
    # children and parents need to be set to managed mode. We do it to make
    # sure to remove the unmanaged flag form the whole tree. The flag could be
    # put there manually. If we didn't do it, the resource may stay unmanaged,
    # as a managed primitive in an unmanaged clone / group is still unmanaged
    # and vice versa.
    res_id = resource_el.attrib["id"]
    parent_el = []
    top_element = find_parent(resource_el, {"resources"})
    if top_element is not None:
        parent_el = cast(
            List[_Element],
            top_element.xpath(
                # a master or a clone which contains a group, a primitive, or a
                # grouped primitive with the specified id
                # OR
                # a group (in a clone, master, etc. - hence //) which contains a
                # primitive with the specified id
                # OR
                # a bundle which contains a primitive with the specified id
                """
                (./master|./clone)[(group|group/primitive|primitive)[@id=$r]]
                |
                //group[primitive[@id=$r]]
                |
                ./bundle[primitive[@id=$r]]
                """,
                r=res_id,
            ),
        )
    children_el = cast(
        List[_Element],
        resource_el.xpath("(./group|./primitive|./group/primitive)"),
    )
    return [resource_el] + parent_el + children_el


def find_resources_to_unmanage(resource_el: _Element) -> List[_Element]:
    """
    Get resources to unmanage to unmanage the specified resource successfully

    resource_el -- resource element
    """
    # resource hierarchy - specified resource - what to return
    # a primitive - the primitive - the primitive
    #
    # a cloned primitive - the primitive - the primitive
    # a cloned primitive - the clone - the primitive
    #   The resource will run on all nodes after unclone. However that doesn't
    #   seem to be bad behavior. Moreover, if monitor operations were disabled,
    #   they wouldn't enable on unclone, but the resource would become managed,
    #   which is definitely bad.
    #
    # a primitive in a group - the primitive - the primitive
    #   Otherwise all primitives in the group would become unmanaged.
    # a primitive in a group - the group - all primitives in the group
    #   If only the group was set to unmanaged, setting any primitive in the
    #   group to managed would set all the primitives in the group to managed.
    #   If the group as well as all its primitives were set to unmanaged, any
    #   primitive added to the group would become unmanaged. This new primitive
    #   would become managed if any original group primitive becomes managed.
    #   Therefore changing one primitive influences another one, which we do
    #   not want to happen.
    #
    # a primitive in a cloned group - the primitive - the primitive
    # a primitive in a cloned group - the group - all primitives in the group
    #   See group notes above
    # a primitive in a cloned group - the clone - all primitives in the group
    #   See clone notes above
    #
    # a bundled primitive - the primitive - the primitive
    # a bundled primitive - the bundle - the bundle and the primitive
    #  We need to unmanage implicit resources create by pacemaker and there is
    #  no other way to do it than unmanage the bundle itself.
    #  Since it is not possible to unbundle a resource, the concers described
    #  at unclone don't apply here. However to prevent future bugs, in case
    #  unbundling becomes possible, we unmanage the primitive as well.
    # an empty bundle - the bundle - the bundle
    #  There is nothing else to unmanage.
    if is_bundle(resource_el):
        in_bundle = get_bundle_inner_resource(resource_el)
        return (
            [resource_el, in_bundle] if in_bundle is not None else [resource_el]
        )
    if is_any_clone(resource_el):
        resource_el = get_clone_inner_resource(resource_el)
    if is_group(resource_el):
        return get_group_inner_resources(resource_el)
    if is_primitive(resource_el):
        return [resource_el]
    return []


def manage(resource_el: _Element, id_provider: IdProvider) -> None:
    """
    Set the resource to be managed by the cluster

    resource_el -- resource element
    """
    nvpair.arrange_first_meta_attributes(
        resource_el,
        {"is-managed": ""},
        id_provider,
    )


def unmanage(resource_el: _Element, id_provider: IdProvider) -> None:
    """
    Set the resource not to be managed by the cluster

    resource_el -- resource element
    """
    nvpair.arrange_first_meta_attributes(
        resource_el,
        {"is-managed": "false"},
        id_provider,
    )


def find_resources_to_delete(resource_el: _Element) -> List[_Element]:
    """
    Get resources to delete, children and parents of the given resource if
    necessary.

    If element is a primitive which is in a clone and you specify one of them,
    you will get elements for both of them. If you specify group element which
    is in a clone then will you get clone, group, and all primitive elements in
    a group and etc.

    resource_el - resource element (bundle, clone, group, primitive)
    """
    result = [resource_el]
    # childrens of bundle, clone, group, clone-with-group
    inner_resource_list = get_inner_resources(resource_el)
    if inner_resource_list:
        result.extend(inner_resource_list)
        inner_resource = inner_resource_list[0]
        if is_group(inner_resource):
            result.extend(get_inner_resources(inner_resource))
    # parents of primitive if needed (group, clone)
    parent_el = get_parent_resource(resource_el)
    if parent_el is None or is_bundle(parent_el):
        return result
    if is_any_clone(parent_el):
        result.insert(0, parent_el)
    if is_group(parent_el):
        group_inner_resources = get_group_inner_resources(parent_el)
        if len(group_inner_resources) <= 1:
            result = [parent_el] + group_inner_resources
            clone_el = get_parent_resource(parent_el)
            if clone_el is not None:
                result.insert(0, clone_el)
    return result


def validate_move(
    resource_element: _Element, master: bool
) -> reports.ReportItemList:
    """
    Validate moving a resource to a node

    resource_element -- the resource to be moved
    master -- limit moving to the master role
    """
    report_list = []
    analysis = _validate_move_ban_clear_analyzer(resource_element)

    if analysis.is_bundle:
        report_list.append(
            ReportItem.error(
                reports.messages.CannotMoveResourceBundle(
                    str(resource_element.get("id", ""))
                )
            )
        )
        return report_list

    if (analysis.is_clone or analysis.is_in_clone) and not (
        analysis.is_promotable_clone or analysis.is_in_promotable_clone
    ):
        report_list.append(
            ReportItem.error(
                reports.messages.CannotMoveResourceClone(
                    str(resource_element.get("id", ""))
                )
            )
        )
        return report_list

    # Allow moving both promoted and demoted instances of promotable clones
    # (analysis.is_promotable_clone). Do not allow to move promoted instances
    # of promotables' inner resources (analysis.is_in_promotable_clone) as that
    # would create a constraint for the promoted role of a group or a primitive
    # which would not make sense if the group or primitive is later moved out
    # of its promotable clone. To be consistent, we do not allow to move
    # demoted instances of promotables' inner resources either. See
    # rhbz#1875301 for details.
    if not master and analysis.is_in_promotable_clone:
        report_list.append(
            ReportItem.error(
                reports.messages.CannotMoveResourcePromotableInner(
                    str(resource_element.get("id", "")),
                    analysis.promotable_clone_id,
                )
            )
        )
    elif master and not analysis.is_promotable_clone:
        report_list.append(
            ReportItem.error(
                reports.messages.CannotMoveResourceMasterResourceNotPromotable(
                    str(resource_element.get("id", "")),
                    promotable_id=analysis.promotable_clone_id,
                )
            )
        )

    return report_list


def validate_ban(
    resource_element: _Element, master: bool
) -> reports.ReportItemList:
    """
    Validate banning a resource on a node

    resource_element -- the resource to be banned
    master -- limit banning to the master role
    """
    report_list = []
    analysis = _validate_move_ban_clear_analyzer(resource_element)

    if master and not analysis.is_promotable_clone:
        report_list.append(
            ReportItem.error(
                reports.messages.CannotBanResourceMasterResourceNotPromotable(
                    str(resource_element.get("id", "")),
                    promotable_id=analysis.promotable_clone_id,
                )
            )
        )

    return report_list


def validate_unmove_unban(
    resource_element: _Element, master: bool
) -> reports.ReportItemList:
    """
    Validate unmoving/unbanning a resource to/on nodes

    resource_element -- the resource to be unmoved/unbanned
    master -- limit unmoving/unbanning to the master role
    """
    report_list = []
    analysis = _validate_move_ban_clear_analyzer(resource_element)

    if master and not analysis.is_promotable_clone:
        report_list.append(
            ReportItem.error(
                reports.messages.CannotUnmoveUnbanResourceMasterResourceNotPromotable(
                    str(resource_element.get("id", "")),
                    promotable_id=analysis.promotable_clone_id,
                )
            )
        )

    return report_list


class _MoveBanClearAnalysis(
    namedtuple(
        "_MoveBanClearAnalysis",
        [
            "is_bundle",
            "is_clone",
            "is_in_clone",
            "is_promotable_clone",
            "is_in_promotable_clone",
            "promotable_clone_id",
        ],
    )
):
    pass


def _validate_move_ban_clear_analyzer(
    resource_element: _Element,
) -> _MoveBanClearAnalysis:
    resource_is_bundle = False
    resource_is_clone = False
    resource_is_in_clone = False
    resource_is_promotable_clone = False
    resource_is_in_promotable_clone = False
    promotable_clone_element = None

    if is_bundle(resource_element):
        resource_is_bundle = True
    elif is_any_clone(resource_element):
        resource_is_clone = True
        if is_master(resource_element) or is_promotable_clone(resource_element):
            resource_is_promotable_clone = True
            promotable_clone_element = resource_element
    elif get_parent_any_clone(resource_element) is not None:
        parent_clone = get_parent_any_clone(resource_element)
        resource_is_in_clone = True
        if parent_clone is not None and (
            is_master(parent_clone) or is_promotable_clone(parent_clone)
        ):
            resource_is_in_promotable_clone = True
            promotable_clone_element = parent_clone
    return _MoveBanClearAnalysis(
        resource_is_bundle,
        resource_is_clone,
        resource_is_in_clone,
        resource_is_promotable_clone,
        resource_is_in_promotable_clone,
        (
            promotable_clone_element.get("id")
            if promotable_clone_element is not None
            else None
        ),
    )



from unittest import TestCase

from pcs.common.pacemaker.resource.operations import CibResourceOperationDto
from pcs.lib.cib.resource import agent
from pcs.lib.resource_agent import (
    ResourceAgentAction,
    ResourceAgentMetadata,
    ResourceAgentName,
)
from pcs.lib.resource_agent.const import OCF_1_0


class GetDefaultOperationInterval(TestCase):
    def test_return_0s_on_name_different_from_monitor(self):
        self.assertEqual("0s", agent.get_default_operation_interval("start"))

    def test_return_60s_on_monitor(self):
        self.assertEqual("60s", agent.get_default_operation_interval("monitor"))


class CompleteOperationsOptions(TestCase):
    def test_add_intervals_everywhere_is_missing(self):
        self.assertEqual(
            agent.complete_operations_options(
                [
                    {"name": "monitor", "interval": "20s"},
                    {"name": "start"},
                ]
            ),
            [
                {"name": "monitor", "interval": "20s"},
                {"name": "start", "interval": "0s"},
            ],
        )


class GetDefaultOperations(TestCase):
    fixture_actions = [
        ResourceAgentAction(
            "custom1", "40s", None, None, None, None, False, False
        ),
        ResourceAgentAction(
            "custom2", "60s", "25s", None, None, None, False, False
        ),
        ResourceAgentAction(
            "meta-data", None, None, None, None, None, False, False
        ),
        ResourceAgentAction(
            "monitor", "30s", "10s", None, None, None, False, False
        ),
        ResourceAgentAction(
            "start", None, "40s", None, None, None, False, False
        ),
        ResourceAgentAction(
            "status", "20s", "15s", None, None, None, False, False
        ),
        ResourceAgentAction(
            "validate-all", None, None, None, None, None, False, False
        ),
    ]
    fixture_actions_meta_only = [
        ResourceAgentAction(
            "meta-data", None, None, None, None, None, False, False
        )
    ]
    maxDiff = None

    @staticmethod
    def fixture_agent(actions):
        return ResourceAgentMetadata(
            ResourceAgentName("ocf", "pacemaker", "Dummy"),
            agent_exists=True,
            ocf_version=OCF_1_0,
            shortdesc="",
            longdesc="",
            parameters=[],
            actions=actions,
        )

    @staticmethod
    def fixture_stonith_agent(actions):
        return ResourceAgentMetadata(
            ResourceAgentName("stonith", None, "fence_test"),
            agent_exists=True,
            ocf_version=OCF_1_0,
            shortdesc="",
            longdesc="",
            parameters=[],
            actions=actions,
        )

    @staticmethod
    def op_fixture(name, interval, timeout):
        return CibResourceOperationDto(
            id="",
            name=name,
            interval=interval,
            description=None,
            start_delay=None,
            interval_origin=None,
            timeout=timeout,
            enabled=None,
            record_pending=None,
            role=None,
            on_fail=None,
            meta_attributes=[],
            instance_attributes=[],
        )

    def test_select_only_actions_for_cib(self):
        self.assertEqual(
            agent.get_default_operations(
                self.fixture_agent(self.fixture_actions)
            ),
            [
                self.op_fixture("custom1", "0s", "40s"),
                self.op_fixture("custom2", "25s", "60s"),
                self.op_fixture("monitor", "10s", "30s"),
                self.op_fixture("start", "40s", None),
            ],
        )

    def test_select_only_actions_for_cib_stonith(self):
        self.assertEqual(
            agent.get_default_operations(
                self.fixture_stonith_agent(self.fixture_actions)
            ),
            [self.op_fixture("monitor", "10s", "30s")],
        )

    def test_select_only_necessary_actions_for_cib(self):
        self.assertEqual(
            agent.get_default_operations(
                self.fixture_agent(self.fixture_actions), necessary_only=True
            ),
            [self.op_fixture("monitor", "10s", "30s")],
        )

    def test_select_only_necessary_actions_for_cib_stonith(self):
        self.assertEqual(
            agent.get_default_operations(
                self.fixture_stonith_agent(self.fixture_actions),
                necessary_only=True,
            ),
            [self.op_fixture("monitor", "10s", "30s")],
        )

    def test_complete_monitor(self):
        self.assertEqual(
            agent.get_default_operations(
                self.fixture_agent(self.fixture_actions_meta_only),
                necessary_only=True,
            ),
            [self.op_fixture("monitor", "60s", None)],
        )

    def test_complete_monitor_stonith(self):
        self.assertEqual(
            agent.get_default_operations(
                self.fixture_stonith_agent(self.fixture_actions_meta_only),
                necessary_only=True,
            ),
            [self.op_fixture("monitor", "60s", None)],
        )



from typing import Optional

ERR_NODE_LIST_AND_ALL_MUTUALLY_EXCLUSIVE = (
    "Cannot specify both --all and a list of nodes."
)
SEE_MAN_CHANGES = "See 'man pcs' -> Changes in pcs-0.10."
HINT_SYNTAX_CHANGE = (
    "Syntax has changed from previous version. " + SEE_MAN_CHANGES
)


def msg_command_replaced(*new_commands):
    new = "', '".join(new_commands)
    return f"This command has been replaced with '{new}'. {SEE_MAN_CHANGES}"


def raise_command_replaced(*new_commands):
    raise CmdLineInputError(message=msg_command_replaced(*new_commands))


class CmdLineInputError(Exception):
    """
    an incorrect command has been entered in the command line
    """

    def __init__(
        self,
        message: Optional[str] = None,
        hint: Optional[str] = None,
        show_both_usage_and_message: bool = False,
    ) -> None:
        """
        message -- explains what was wrong with the entered command
        hint -- provides an additional hint how to proceed
        show_both_usage_and_message -- show both the message and usage

        The routine which handles this exception behaves according to whether
        the message was specified (prints this message to user) or not (prints
        appropriate part of documentation). If show_both_usage_and_message is
        True, documentation will be printed first and the message will be
        printed after that. Hint is printed every time as the last item.
        """
        super().__init__(message)
        self.message = message
        self.hint = hint
        self.show_both_usage_and_message = show_both_usage_and_message



from unittest import TestCase

from pcs.common import reports
from pcs.lib.commands import constraint

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools


class SetCreate(TestCase):
    def setUp(self):
        resources_xml = """
            <resources>
                <group id="grAB">
                    <primitive class="ocf" id="A" provider="pacemaker" type="Dummy"/>
                    <primitive class="ocf" id="B" provider="pacemaker" type="Dummy"/>
                </group>
                <group id="grC">
                    <primitive class="ocf" id="C" provider="pacemaker" type="Dummy"/>
                </group>
                <primitive class="ocf" id="D" provider="pacemaker" type="Dummy"/>
            </resources>
        """
        self.env_assist, self.config = get_env_tools(self)
        self.config.runner.cib.load(resources=resources_xml)

    def test_deny_resources_from_one_group_in_one_set(self):
        self.env_assist.assert_raise_library_error(
            lambda: constraint.order.create_with_set(
                self.env_assist.get_env(),
                [{"ids": ["A", "B"], "options": {}}],
                {},
            ),
            [
                fixture.error(
                    reports.codes.CANNOT_SET_ORDER_CONSTRAINTS_FOR_RESOURCES_IN_THE_SAME_GROUP,
                )
            ],
            expected_in_processor=False,
        )

    def test_deny_resources_from_one_group_in_different_sets(self):
        self.env_assist.assert_raise_library_error(
            lambda: constraint.order.create_with_set(
                self.env_assist.get_env(),
                [{"ids": ["A"], "options": {}}, {"ids": ["B"], "options": {}}],
                {},
            ),
            [
                fixture.error(
                    reports.codes.CANNOT_SET_ORDER_CONSTRAINTS_FOR_RESOURCES_IN_THE_SAME_GROUP,
                )
            ],
            expected_in_processor=False,
        )

    def test_allow_resources_from_different_groups(self):
        constraints_xml = """
            <constraints>
                <rsc_order id="order_set_AACCAA">
                    <resource_set id="order_set_AACCAA_set">
                        <resource_ref id="A"/>
                        <resource_ref id="C"/>
                    </resource_set>
                    <resource_set id="order_set_AACCAA_set-1">
                        <resource_ref id="A"/>
                        <resource_ref id="D"/>
                    </resource_set>
                </rsc_order>
            </constraints>
        """
        self.config.env.push_cib(constraints=constraints_xml)

        constraint.order.create_with_set(
            self.env_assist.get_env(),
            [
                {"ids": ["A", "C"], "options": {}},
                {"ids": ["A", "D"], "options": {}},
            ],
            {},
        )



from unittest import TestCase

from pcs.lib import resource_agent as ra


class ResourceAgentName(TestCase):
    def test_full_name_3_parts(self):
        self.assertEqual(
            ra.ResourceAgentName("standard", "provider", "type").full_name,
            "standard:provider:type",
        )

    def test_full_name_2_parts_none(self):
        self.assertEqual(
            ra.ResourceAgentName("standard", None, "type").full_name,
            "standard:type",
        )

    def test_full_name_2_parts_empty(self):
        self.assertEqual(
            ra.ResourceAgentName("standard", "", "type").full_name,
            "standard:type",
        )

    def test_is_stonith_yes(self):
        self.assertTrue(
            ra.ResourceAgentName("stonith", "pacemaker", "Dummy").is_stonith
        )

    def test_is_stonith_no(self):
        self.assertFalse(
            ra.ResourceAgentName("lsb", None, "fence_xvm").is_stonith
        )

    def test_is_fake_pcmk_agent_yes(self):
        self.assertTrue(
            ra.ResourceAgentName(
                ra.const.FAKE_AGENT_STANDARD, None, "Dummy"
            ).is_pcmk_fake_agent
        )

    def test_is_fake_pcmk_agent_no(self):
        self.assertFalse(
            ra.ResourceAgentName("pacemaker", None, "fenced").is_pcmk_fake_agent
        )


class ProvidesUnfencing(TestCase):
    @staticmethod
    def _fixture_metadata(name, actions):
        return ra.ResourceAgentMetadata(
            name,
            agent_exists=True,
            ocf_version=ra.const.OCF_1_0,
            shortdesc=None,
            longdesc=None,
            parameters=[],
            actions=actions,
        )

    @staticmethod
    def _fixture_action(name, automatic, on_target):
        return ra.ResourceAgentAction(
            name=name,
            timeout=None,
            interval=None,
            role=None,
            start_delay=None,
            depth=None,
            automatic=automatic,
            on_target=on_target,
        )

    def test_not_stonith(self):
        self.assertFalse(
            self._fixture_metadata(
                ra.ResourceAgentName("ocf", "pacemaker", "Dummy"),
                [self._fixture_action("on", True, True)],
            ).provides_unfencing
        )

    def test_not_automatic(self):
        self.assertFalse(
            self._fixture_metadata(
                ra.ResourceAgentName("stonith", None, "fence_xvm"),
                [self._fixture_action("on", False, True)],
            ).provides_unfencing
        )

    def test_not_on_target(self):
        self.assertFalse(
            self._fixture_metadata(
                ra.ResourceAgentName("stonith", None, "fence_xvm"),
                [self._fixture_action("on", True, False)],
            ).provides_unfencing
        )

    def test_not_action_on(self):
        self.assertFalse(
            self._fixture_metadata(
                ra.ResourceAgentName("stonith", None, "fence_xvm"),
                [self._fixture_action("off", True, True)],
            ).provides_unfencing
        )

    def test_true(self):
        self.assertTrue(
            self._fixture_metadata(
                ra.ResourceAgentName("stonith", None, "fence_xvm"),
                [
                    self._fixture_action("on", False, True),
                    self._fixture_action("on", True, False),
                    self._fixture_action("off", True, True),
                    self._fixture_action("on", True, True),
                ],
            ).provides_unfencing
        )


class UniqueParameterGroups(TestCase):
    @staticmethod
    def _fixture_metadata(parameters):
        return ra.ResourceAgentMetadata(
            ra.ResourceAgentName("ocf", "pacemaker", "Dummy"),
            agent_exists=True,
            ocf_version=ra.const.OCF_1_0,
            shortdesc=None,
            longdesc=None,
            parameters=parameters,
            actions=[],
        )

    @staticmethod
    def _fixture_parameter(name, unique_group):
        return ra.ResourceAgentParameter(
            name,
            shortdesc=None,
            longdesc=None,
            type="string",
            default=None,
            enum_values=None,
            required=False,
            advanced=False,
            deprecated=False,
            deprecated_by=[],
            deprecated_desc=None,
            unique_group=unique_group,
            reloadable=False,
        )

    def test_no_groups(self):
        self.assertEqual(
            self._fixture_metadata(
                [
                    self._fixture_parameter("param_1", None),
                    self._fixture_parameter("param_2", None),
                    self._fixture_parameter("param_3", None),
                ]
            ).unique_parameter_groups,
            {},
        )

    def test_groups(self):
        self.assertEqual(
            self._fixture_metadata(
                [
                    self._fixture_parameter("param_1", None),
                    self._fixture_parameter("param_2", "group_A"),
                    self._fixture_parameter("param_3", None),
                    self._fixture_parameter("param_4", ""),
                    self._fixture_parameter("param_5", "group_B"),
                    self._fixture_parameter("param_6", "group_A"),
                ]
            ).unique_parameter_groups,
            {
                "group_A": {"param_2", "param_6"},
                "group_B": {"param_5"},
            },
        )



import json

from pcs.common import communication
from pcs.common.interface.dto import to_dict

from pcs_test.tools.command_env.mock_node_communicator import (
    place_communication,
)


class ScsiShortcuts:
    def __init__(self, calls):
        self.__calls = calls

    def unfence_node(
        self,
        original_devices=(),
        updated_devices=(),
        node_labels=None,
        communication_list=None,
        name="http.scsi.unfence_node",
    ):
        """
        Create a calls for node unfencing

        list original_devices -- list of scsi devices before an update
        list updated_devices -- list of scsi devices after an update
        list node_labels -- create success responses from these nodes
        list communication_list -- use these custom responses
        string name -- the key of this call
        """
        if (node_labels is None and communication_list is None) or (
            node_labels and communication_list
        ):
            raise AssertionError(
                "Exactly one of 'node_labels', 'communication_list' "
                "must be specified"
            )

        if node_labels:
            communication_list = [
                dict(
                    label=node,
                    raw_data=json.dumps(
                        dict(
                            node=node,
                            original_devices=original_devices,
                            updated_devices=updated_devices,
                        )
                    ),
                )
                for node in node_labels
            ]
        place_communication(
            self.__calls,
            name,
            communication_list,
            action="api/v1/scsi-unfence-node/v2",
            output=json.dumps(
                to_dict(
                    communication.dto.InternalCommunicationResultDto(
                        status=communication.const.COM_STATUS_SUCCESS,
                        status_msg=None,
                        report_list=[],
                        data=None,
                    )
                )
            ),
        )

    def unfence_node_mpath(
        self,
        node_key_map,
        original_devices=(),
        updated_devices=(),
        node_labels=None,
        communication_list=None,
        name="http.scsi.unfence_node",
    ):
        """
        Create a calls for node unfencing

        dict node_key_map -- map of node name to its registration key
        list original_devices -- list of scsi devices before an update
        list updated_devices -- list of scsi devices after an update
        list node_labels -- create success responses from these nodes
        list communication_list -- use these custom responses
        string name -- the key of this call
        """
        if (node_labels is None and communication_list is None) or (
            node_labels and communication_list
        ):
            raise AssertionError(
                "Exactly one of 'node_labels', 'communication_list' "
                "must be specified"
            )

        if node_labels:
            communication_list = [
                dict(
                    label=node,
                    raw_data=json.dumps(
                        dict(
                            key=node_key_map[node],
                            original_devices=original_devices,
                            updated_devices=updated_devices,
                        )
                    ),
                )
                for node in node_labels
            ]
        place_communication(
            self.__calls,
            name,
            communication_list,
            action="api/v1/scsi-unfence-node-mpath/v1",
            output=json.dumps(
                to_dict(
                    communication.dto.InternalCommunicationResultDto(
                        status=communication.const.COM_STATUS_SUCCESS,
                        status_msg=None,
                        report_list=[],
                        data=None,
                    )
                )
            ),
        )
